"""Logic for updating QPForm.

Functions:
update_qpform_all
find_subnet_node
find_buffer
locate_rivers
load_utility
load_market
load_tradepoint
load_piecewise
remove_segments
state_limits
load_storage
load_hydro_storage
load_ac_dc
fit_fcn
fit_with_intercept
fit_residual
fit_coproduction
fit_fcn_2
fit_with_intercept2
max_ac_dc_gen
update_utility_sellback_limit
load_solar
"""

import numpy as np

from eagers.config.network import NETWORK_NAME_ABBR_MAP, OUTPUT_NETWORK_MAP
from eagers.basic.ss_response import ss_response
from eagers.solver.ecos_ls import ecos_ls


def update_qpform_all(generator, subnet, test_data, scale_time):
    """Loads generators for economic dispatch.
    This function identifies the values that will be used to represent each
    generator in the quadratic optimizations.
    """
    n_g = len(generator)
    n = 5  # Number of segments in piecewise quadratic fits.
    qpform = [{}] * n_g
    for i, gen in enumerate(generator):
        if gen['_type'] in ['RTUtilityElectric','RTUtilityDiesel','RTUtilityCool','RTUtilityHeat','RTUtilityNatgas','TSUtilityElectric','TSUtilityDiesel','TSUtilityCool','TSUtilityHeat','TSUtilityNatgas']:
            qpform[i] = load_utility(gen)
        elif gen['_type'] == 'Tradepoint':
            qpform[i] = load_tradepoint(gen)
        elif gen['_type'] == 'Market':
            qpform[i] = load_market(gen)
        elif gen['_type'] in ['ElecGeneratorAC', 'ElecGeneratorDC','CombHeatPowerAC', 'CombHeatPowerDC', 'Chiller',
                'Heater', 'CoolingTower', 'Electrolyzer', 'HydrogenGenerator']:
            max_seg = max([1,min([n,len(gen['output']['capacity'])-3])])
            qpform[i] = load_piecewise(gen, max_seg)
            if gen['_type'] in ['ElecGeneratorAC', 'CombHeatPowerAC']:
                gen['lower_bound'] = gen['startup']['electricity'][-1]
            elif gen['_type'] in ['ElecGeneratorDC', 'CombHeatPowerDC']:
                gen['lower_bound'] = gen['startup']['direct_current'][-1]
            elif gen['_type'] in ['Chiller']:
                gen['lower_bound'] = gen['startup']['cooling'][-1]
            elif gen['_type'] in ['Heater']:
                gen['lower_bound'] = gen['startup']['heat'][-1]
            elif gen['_type'] in ['CoolingTower']:
                gen['lower_bound'] = gen['startup']['heat_reject'][-1]
            elif gen['_type'] in ['Electrolyzer', 'HydrogenGenerator']:
                gen['lower_bound'] = gen['startup']['hydrogen'][-1]
            ramp_rate = ss_response(gen['size_kw'], gen['lower_bound'],gen['ramp_rate'],gen['state_space']['a'],gen['state_space']['b'],gen['state_space']['c'],gen['state_space']['d'])
            ramp_rate = ramp_rate/scale_time
            qpform[i]['ramp'] = {}
            qpform[i]['ramp']['b'] = [ramp_rate for j in range(2)] # [ramp up, ramp down]
        elif gen['_type'] == 'Solar':
            qpform[i] = load_solar(gen,subnet)
        elif gen['_type'] == 'HydroStorage':
            qpform[i] = load_hydro_storage(gen, scale_time)
            qpform[i]['ramp'] = [gen['ramp_up'] / scale_time,gen['ramp_up'] / scale_time]
        elif gen['_type'] in ['ElectricStorage','ThermalStorage','HydrogenStorage']:
            qpform[i] = load_storage(gen, scale_time)
            if gen['_type'] == 'ElectricStorage' and not 'direct_current' in subnet:
                qpform[i]['output'] = {}
                qpform[i]['output']['e'] = [[1]]
        elif gen['_type'] == 'ACDCConverter':
            qpform[i] = load_ac_dc(gen)        
        qpform[i]['type'] = assign_type(gen)
        qpform[i]['name'] = gen['name']
        if 'source' in gen:
            qpform[i]['source'] = gen['source']
        qpform[i]['enabled'] = True
        

    max_ac_gen, max_dc_gen = max_ac_dc_gen(subnet,qpform)
    qpform = update_utility_sellback_limit(qpform,test_data,max_ac_gen, max_dc_gen)
    qpform = locate_rivers(subnet,qpform)
    qpform = find_buffer(subnet,qpform)
    qpform = find_subnet_node(qpform,subnet)
    return qpform

def assign_type(gen):
    # Assign generator type.
    gen_type = None
    if gen['_type'] in ['CombHeatPowerAC', 'CombHeatPowerDC']:
        gen_type = 'CombinedHeatPower'
    elif gen['_type'] in ['ElecGeneratorAC', 'ElecGeneratorDC']:
        gen_type = 'ElectricGenerator'
    elif gen['_type'] in ['RTUtilityCool','RTUtilityDiesel','RTUtilityElectric','RTUtilityHeat', 'RTUtilityNatgas', 'TSUtilityCool','TSUtilityDiesel','TSUtilityElectric','TSUtilityHeat', 'TSUtilityNatgas']:
        gen_type = 'Utility'
    else:
        gen_type = gen['_type']
    return gen_type

def find_subnet_node(qpform,subnet):
    gen_names = [qpform[i]['name'] for i in range(len(qpform))]
    for i in range(len(qpform)):
        qpform[i]['subnet_node'] = {}
    for net in subnet['network_names']:
        for n in range(len(subnet[net]['nodes'])):
            equip= subnet[net]['equipment'][n]
            for j in equip:
                gen_i = gen_names.index(j)
                qpform[gen_i]['subnet_node'][net] = n
    return qpform

def find_buffer(subnet,qpform):
    """calculate the upper and lower buffer thresholds for each storage system
    """
    gen_names = [qpform[i]['name'] for i in range(len(qpform))]
    for net in subnet['network_names']:
        for m in range(len(subnet[net]['nodes'])):
            equip = subnet[net]['equipment'][m]
            for j in equip:
                k = gen_names.index(j)
                if 'stor' in qpform[k] and 'u' in qpform[k]['states'][0]:
                    if 'buffer' in qpform[k]:
                        buff_perc = qpform[k]['buffer']
                    else:
                        buff_perc = 0
                    if net =='hydro':
                        dischargeCapacity = (qpform[k]['max_gen_flow'] + qpform[k]['s']['ub'])*24/12.1 #amount the resevoir can discharge in a day #flow rate in 1000 ft^3 converted to 1000 acre ft (1000 acre-ft = 12.1 x 1000 ft^3/s * 1 hr)
                        buffer = min([(buff_perc/100)*qpform[k]['stor']['usable_size'],dischargeCapacity])
                    else:
                        buffer = (buff_perc/100)*qpform[k]['stor']['usable_size']
                    qpform[k]['link']['bineq'][-2] = -buffer
                    qpform[k]['link']['bineq'][-1] = qpform[k]['stor']['usable_size'] - buffer
                    qpform[k]['u']['ub'] = buffer
                    qpform[k]['l']['ub'] = buffer           
    return qpform

def locate_rivers(subnet,qpform):
    gen_names = [qpform[i]['name'] for i in range(len(qpform))]
    if 'hydro' in subnet:
        nhn = len(subnet['hydro']['nodes'])    
        for n in range(nhn):
            equip = subnet['hydro']['equipment'][n]
            for j in equip:
                gen_i = gen_names.index(j)
                if 'w' in qpform[gen_i]['output']:
                    qpform[gen_i]['subnet_node']['hydro'] = n
                    if qpform[gen_i]['type'] == 'HydroStorage':
                        qpform[gen_i]['hydro']['downriver'] = subnet['hydro']['line_number'][n][0]
    return qpform

def load_utility(gen):
    qpform = {}
    if 'Electric' in gen['_type']:
        gen['source'] = 'electricity'
    elif 'Heat' in gen['_type']:
        gen['source'] = 'heat'
    elif 'Cool' in gen['_type']:
        gen['source'] ='cooling'
    elif 'Coal' in gen['_type']:
        gen['source'] = 'coal'
    elif 'Natgas' in gen['_type']:
        gen['source'] ='ng'
    elif 'Diesel' in gen['_type']:
        gen['source'] = 'diesel'
    else:
        gen['source'] ='nuclear'
    qpform['source'] = gen['source']
    if not gen['source'] in ['ng', 'diesel','coal','nuclear']:# gen is not a fuel utility.
        # No sellback allowed (only 1 state).
        qpform['output'] = {}
        qpform['states'] = [['x']]
        qpform['x'] = {}
        qpform['x']['h'] = 0
        qpform['x']['f'] = 1
        
        qpform['x']['ub'] = float('inf')
        qpform['ub'] = float('inf')
        if gen['source'] == 'electricity':
            s = 'e'
        elif gen['source'] == 'heat':
            s = 'h'# Loads the parameters for a distric heating supply.
        elif gen['source'] == 'cooling':
            s = 'c'# Loads the parameters for a distric cooling supply.
        qpform['output'][s] = [[1]]
        if 'min_import_thresh' in gen:
            qpform['x']['lb'] = gen['min_import_thresh']
            qpform['sellback_rate'] = gen['sellback_rate']
            qpform['sellback_perc'] = gen['sellback_perc']
            if gen['min_import_thresh'] <= 0 and (gen['sellback_rate'] > 0  or (gen['sellback_rate'] == -1 and gen['sellback_perc'] != 1)):
                # Add sellback state.
                qpform['states'] = [['x','y']]
                qpform['y'] = {}
                if gen['sellback_rate'] > 0:
                    qpform['y']['f'] = -gen['sellback_rate']  # Constant sellback rate
                else:
                    # Ensure less than 1, so no issues with pass-through power.
                    qpform['y']['f'] =-min(gen['sellback_perc']/100, 1 - 1e-6)
                qpform['y']['h'] = 0
                qpform['y']['lb'] = 0
                qpform['y']['ub'] = float('inf')
                qpform['x']['lb'] = 0
                qpform['output'][s][0].append(-1)
        else:
            qpform['x']['lb'] = 0
    else:
        qpform['states'] = [[]]
        qpform['output'] = []
        qpform['ub'] = float('inf')
        #TODO timestamp and rate?
    if 'RTUtility' in gen['_type']:
        kys = ['sum_start_month','sum_start_day','win_start_month','win_start_day','sum_rate_table','win_rate_table','sum_rates','win_rates']
    elif 'TSUtility' in gen['_type']:
        kys = ['timestamp','rate']
    for k in kys:
        qpform[k] = gen[k]
    return qpform

def load_market(market):
    qpform = {}
    qpform['states'] = [['x']]
    qpform['x']['h'] = 0
    qpform['x']['f'] = 1
    qpform['x']['lb'] = 0 #initial market assumes 0 bids
    qpform['x']['ub'] = 0 # upper and lower bound set to the awarded bid
    qpform['output']['e'] = [[1]]
    if market.MinImportThresh<=0: #add sell back state
        qpform['states'] = [['x','y']]
        #Value for selling to market will be updated as market closes
        if 'sellback_rate' in market:
            qpform['y']['f'] = -market.sellback_rate #default selling rate set to -1 
        else:
            qpform['y']['f'] = -1 #set to sellback indicated by market
        qpform['y']['h'] = 0
        qpform['y']['lb'] = 0 #initial market assumes 0 bids
        qpform['y']['ub'] = 0 #upper and lower bound set to the awarded bid
        qpform['output']['e'][0].append(-1)
    return qpform

def load_tradepoint(t_point):
    qpform = {}
    qpform['six_param'] = [t_point.b0,t_point.b1,t_point.max_buy,t_point.s0,t_point.s1,t_point.max_sell]
    qpform['states'] = [['x','y']]
    qpform['x']['f'] = t_point.b0
    qpform['x']['h'] = t_point.b1
    qpform['x']['lb'] = 0
    qpform['x']['ub'] = t_point.max_buy

    qpform['y']['f'] = t_point.s0 #selling rate
    qpform['y']['h']= t_point.s1
    qpform['y']['lb'] = 0
    qpform['y']['ub'] = t_point.max_sell
    qpform['output']['e'] = [[1,-1]]
    qpform['lb'] = qpform['y']['ub']
    qpform['ub'] = qpform['x']['ub']
    return qpform 

def load_piecewise(gen, n):
    """This function loads the parameters for a combined heat and power
    generator, regular electric generator, or chiller.\n
    n is number of segments.\n
    order is either 1 or 2. When order is 2 it solves for the quadratic.\n
    Coefficients of C = c_0 + a_1*x_1 + a_2*x_2 + ... + a_n*X_n + b_1*x_1^2
        + b_2*x_2^2 + ... + b_n*X_n^2 subject to b_i > 0, and a_i > a_(i-1)
        + b_(i-1)*(x_i)_max\n
    If order is 1 it solves for linear coefficients of
        C = c_0 + a_1*x_1 + a_2*x_2 + ... + a_n*X_n subject to a_i > a_(i-1)"""
    qpform = {}
    qpform['output'] ={}
    order = 2
    if gen['_type'] == 'Electrolyzer':
        lb = gen['startup']['hydrogen'][0,-1]
        qpform['output']['hy'] = [[1]]
        efficiency = gen['output']['hydrogen']
    elif gen['_type'] in ['CombHeatPowerAC', 'CombHeatPowerDC', 'ElecGeneratorAC', 'ElecGeneratorDC', 'HydrogenGenerator']:
        if 'electricity' in gen['output']:
            lb = gen['startup']['electricity'][-1]
            efficiency = gen['output']['electricity']
            qpform['output']['e'] = [[1]]
        elif 'direct_current' in gen['output']:
            lb = gen['startup']['direct_current'][-1]
            efficiency = gen['output']['direct_current']
            qpform['output']['dc'] = [[1]]
    elif gen['_type'] == 'Heater':
        lb = gen['startup']['heat'][-1]
        qpform['output']['h'] = [[1]]
        efficiency = gen['output']['heat']
    elif gen['_type'] == 'Chiller':
        order = 1
        lb = gen['startup']['cooling'][-1]
        qpform['output']['c'] = [[1]]
        efficiency = gen['output']['cooling']
    elif gen['_type'] == 'CoolingTower':
        order = 1
        lb = gen['startup']['heat_reject'][-1]
        qpform['output']['cw'] = [[-1]]
        efficiency = gen['output']['heat_reject']
    
    qpform['max_eff'] = max(efficiency)
    p_i = max(range(len(efficiency)), key=efficiency.__getitem__)
    qpform['max_eff_point'] = gen['output']['capacity'][p_i]*gen['size_kw']

    ub = gen['size_kw']
    qpform['lb'] = lb
    qpform['ub'] = ub
    operation_range = [i for i, x in enumerate(gen['output']['capacity'])
        if float('%.6f' % x) >= float('%.6f' % (lb/ub))]
    # Sort list of (index, value) pairs based on values, where values are the
    # elements of capacity at indices specified by operation_range.
    valid_capacity = [gen['output']['capacity'][i] for i in operation_range]
    valid_eff = [efficiency[i] for i in operation_range]
    c_sort = sorted(range(len(valid_capacity)), key=lambda k: valid_capacity[k])
    p = [valid_capacity[i] for i in c_sort]
    eff = [valid_eff[i] for i in c_sort]
    
    qpform['eff'] = eff
    qpform['cap'] = [p[i]*ub for i in range(len(p))]
    if gen['_type'] in ['CombHeatPowerAC', 'CombHeatPowerDC']:
        chp_heat_eff = gen['output']['heat']
        valid_chp_eff = [chp_heat_eff[i] for i in operation_range]
        qpform['chp_eff'] = [valid_chp_eff[i] for i in c_sort]
    # Cost of generator in terms of input at outputs p.
    y = []
    for i in range(len(eff)):
        if eff[i]>0:
            y.append(p[i]/eff[i])
        else:
            y.append(0)

    seg_end = [lb/ub + (1-lb/ub)*(i+1)/n for i in range(n)]
    _, a, _ = fit_fcn(p, y, seg_end, n, order, 0)  # In fit A: c_0 = 0
    _, b, c_0 = fit_fcn(p, y, seg_end, n, order, 1)

    
    qpform['const_cost'] = c_0 * ub
    if 'start_cost' in gen:
        qpform['start_cost'] = gen['start_cost']
    else: 
        qpform['start_cost'] = 0
    qpform['const_demand'] = {}
    x_max = [seg_end[0]]
    x_max.extend([seg_end[i+1] - seg_end[i] for i in range(n-1)])

    if gen['_type'] in ['CombHeatPowerAC', 'CombHeatPowerDC']:
        h_0, heat_out, qpform['max_heat'] = fit_coproduction(gen, a, b, n)
        qpform['const_demand']['district_heat'] = -h_0 * ub
    else:
        heat_out = [None,None]

    a, keep_a = remove_segments(a, n, x_max, order, heat_out[0])
    b, keep_b = remove_segments(b, n, x_max, order, heat_out[1])
    # n_a, n_b must be integers for later use in setting list length
    n_a = 0
    for i in range(n):
        if keep_a[i]:
            n_a += 1
    n_b = 0
    for i in range(n):
        if keep_b[i]:
            n_b += 1
    qpform['states'] = [[],[]]
    if gen['_type'] == 'Chiller':
        net = OUTPUT_NETWORK_MAP[gen['source']]
        source = NETWORK_NAME_ABBR_MAP[net]
        if source == 'h':
            # Absorption chiller has no direct electrical load, but later pump for cooling tower can be added.
            qpform['output']['e'] = [[0]]
        qpform['output'][source] = [[] for j in range(2)]
        qpform['output'][source][0] = [-a[j] for j in range(len(a)) if keep_a[j]]
        qpform['output'][source][1] = [-b[j] for j in range(len(b)) if keep_b[j]]
        qpform['output']['cw'] = [[] for j in range(2)]
        qpform['output']['cw'][0] = [1-qpform['output'][source][0][i] for i in range(n_a)]
        qpform['output']['cw'][1] = [1-qpform['output'][source][1][i] for i in range(n_b)]
        qpform['const_demand'][net] = c_0 * ub
        qpform['const_demand']['cooling_water'] = -c_0 * ub# Heat added to water loop appears as a negative demand for energy in the water loop.
        qpform = state_limits(qpform,[0 for i in range(len(a))],0,ub,x_max,keep_a,order,0)
        qpform = state_limits(qpform,[0 for i in range(len(b))],lb,ub,x_max,keep_b,order,1)
        qpform['const_cost'] = 0
    elif gen['_type'] == 'CoolingTower':
        qpform['output']['e'] = [[] for j in range(2)]
        qpform['output']['e'][0] = [-a[j] for j in range(len(a)) if keep_a[j]]
        qpform['output']['e'][1] = [-b[j] for j in range(len(b)) if keep_b[j]]
        qpform['const_demand']['electrical'] = c_0 * ub
        qpform['const_demand']['cooling_water'] = -c_0 * ub# Heat added to water loop appears as a negative demand for energy in the water loop.
        qpform = state_limits(qpform,[0 for i in range(len(a))],0,ub,x_max,keep_a,order,0)
        qpform = state_limits(qpform,[0 for i in range(len(b))],lb,ub,x_max,keep_b,order,1)
        qpform['const_cost'] = 0
    elif gen['_type'] in ['CombHeatPowerAC', 'CombHeatPowerDC']:
        qpform = state_limits(qpform,a,0,ub,x_max,keep_a,order,0)
        qpform = state_limits(qpform,b,lb,ub,x_max,keep_b,order,1)
        qpform['output']['h'] = [[0 for i in range(n)] for j in range(2)]
        qpform['output']['h'][0][:n_a] = [heat_out[0][i] for i in range(n) if keep_a[i]]
        qpform['output']['h'][1][:n_b] = [heat_out[1][i] for i in range(n) if keep_b[i]]
    else:
        qpform = state_limits(qpform,a,0,ub,x_max,keep_a,order,0)
        qpform = state_limits(qpform,b,lb,ub,x_max,keep_b,order,1)
    return qpform
    
def remove_segments(a, n, x_max, order, co_prod):
    """Remove segments that that have less than 1% deviation in slope 
    (and co-production if applicable) to previous segment"""
    tol = 1e-2
    keep = [True for i in range(order*n)]
    for i in range(1, n):
        if (((a[i]==0 and a[i-1]==0) and (a[i-1]!=0 and abs((a[i] - a[i-1])/a[i-1]) < tol))
                and (order == 1 or a[n+i]*x_max[i] < tol*a[i-1])
                and (co_prod is None or abs((co_prod[i] - co_prod[i-1])/co_prod[i-1]) < tol)):
            keep[i] = False
            if order == 2:
                keep[n+i] = False
    for i in range(n):
        if order == 2 and a[n+i]*x_max[i] < 0.1*tol*a[i]:
            a[n+i] = 0
    return a, keep

def state_limits(qpform,fit_terms,lower_bound,upper_bound,x_max,keep,order,fit):
    ## put properteries of cost and limits into states A, B, ...
    letters = ['a','b','c','d','e','f','g','h','i','j','k','l','m','n']
    n_states = 0
    for i in range(len(keep)):
        if keep[i]:
            n_states +=1/order
    qpform['states'][fit] = letters[:int(n_states)]
    j = -1
    for i in range(len(x_max)):
        if keep[i]:    
            j +=1
            if not letters[j] in qpform:
                qpform[letters[j]] = {}
                qpform[letters[j]]['h'] = [[],[]]
                qpform[letters[j]]['f'] = [[],[]]
                qpform[letters[j]]['lb'] = [[],[]]
                qpform[letters[j]]['ub'] = [[],[]]
            if order == 2:
                qpform[letters[j]]['h'][fit] = fit_terms[len(x_max)+i]/upper_bound ## scale by size of gen (later can normalize by changing QPform.Output so that all states are 0-1)
            else:
                qpform[letters[j]]['h'][fit] = 0
            qpform[letters[j]]['f'][fit] = fit_terms[i]
            if i ==0:
                qpform[letters[j]]['lb'][fit] = lower_bound
                qpform[letters[j]]['ub'][fit] = max([lower_bound,x_max[i]*upper_bound])
            else:
                qpform[letters[j]]['lb'][fit] = 0
                qpform[letters[j]]['ub'][fit] = x_max[i]*upper_bound
        else:
            qpform[letters[j]]['ub'][fit] += x_max[i]*upper_bound
    return qpform

def load_storage(gen, scale_time):
    """This function just directs to either hot or cold thermal storage.\n
    If we can get rid of the CS, HS structures then this function can be
    eliminated, because all storage can be handled the same way."""

    qpform = {}
    stor = {}
    qpform['output'] = {}

    stor['size'] = gen['size_kwh'] * scale_time
    # Usable size (state x is 0 -> usable size, must calculate this
    # initial condition each time from capacity & SOC).
    stor['usable_size'] = gen['size_kwh'] * scale_time
    if 'max_dod' in gen:
        stor['usable_size'] = stor['usable_size'] * gen['max_dod'] / 100 
    # Self-discharge per hour (fraction of total charge).
    stor['self_discharge'] = gen['self_discharge']
    if 'fill_rate' in gen:
        # Thermal or hydrogen storage.
        stor['peak_disch'] = gen['disch_rate_perc'] / 100 * gen['size_kwh']  # [Thermal kW out]
        stor['peak_charge'] = gen['fill_rate_perc'] / 100 * gen['size_kwh']  # [Thermal kW in]
        stor['charge_eff'] = gen['charge_eff']
        stor['disch_eff'] = gen['disch_eff']
        if gen['source'] == 'cooling':
            qpform['output']['c'] = [[1]]
        elif gen['source'] == 'heat':
            qpform['output']['h'] = [[1]]
        elif gen['source'] == 'hydrogen':
            qpform['output']['hy'] = [[1]]
    else:
        # Electric battery.
        qpform['output']['dc'] = [[1]]
        stor['voltage'] = gen['voltage']
        disch_current = gen['peak_disch'] * gen['size_kwh'] / stor['voltage'] * 1000
        # Scale so the loss of power is equivelant to that specified at 100 Amps.
        stor['disch_resist_scaled'] = (100 / disch_current) * gen['disch_resist'] / 1000
        # Keep in mind when calculating loss as function of discharge current.
        disch_volt_loss = disch_current * stor['disch_resist_scaled']
        charge_current = gen['peak_charge'] * gen['size_kwh'] / stor['voltage'] * 1000
        # Scale so the loss of power is equivelant to that specified at 10 Amps.
        stor['charge_resist_scaled'] = (100/charge_current) * gen['charge_resist'] / 1000
        charge_volt_loss = charge_current * stor['charge_resist_scaled']
        stor['charge_eff'] = stor['voltage'] / (stor['voltage'] + charge_volt_loss)
        stor['disch_eff'] = (stor['voltage'] - disch_volt_loss) / stor['voltage']
        stor['peak_disch'] = disch_current * stor['voltage'] * stor['disch_eff'] / 1000 # Peak discharge power [kW out].
        stor['peak_charge'] = charge_current * stor['voltage'] / stor['charge_eff'] / 1000  # Peak charge power [kW in].
    qpform['stor'] = stor
    # qpform['ramp_rate'] = [stor['peak_disch'],stor['peak_charge']]  # Storage discharge constraint [kW].
    qpform['states'] = [['x']] # State of charge, charging power, no buffers.
    qpform['x'] = {}
    qpform['x']['lb'] = 0
    qpform['x']['ub'] = stor['usable_size']
    qpform['x']['h'] = 0
    qpform['x']['f'] = 0
    a = 1 / stor['charge_eff'] - stor['disch_eff']
    qpform['ramp'] = {}
    qpform['ramp']['b'] = [stor['peak_charge'],stor['peak_disch']]
    qpform['link'] = {}
        # SOC2-SOC1 < peak_charge; SOC1-SOC2 < peak_disch
    if a != 0:# Not ideal storage, add charging state.
        qpform['states'] = [['x','y']] # State of charge, charging power, no buffers.
        qpform['link']['ineq'] = [[a, -1]] # (1/nc - nd) * (SOC(t) - SOC(t-1)) - charging <------ charging is 1/inefficiency + 1
        qpform['link']['bineq'] = [0]
        qpform['y'] = {}
        qpform['y']['lb'] = 0
        qpform['y']['ub'] = stor['peak_charge']  # The limit on how much charging power can be delivered is handled by the generators' limits, so put inf here to prevent redundancy.
        qpform['y']['h'] = 0  # Cost of charging power is handled by the generators.
        qpform['y']['f'] = 0
    if 'buffer' in gen and gen['buffer'] != 0:  # Buffer states.
        if a == 0: # Ideal storage, ignore charging state.
            qpform['states'] = [['x','u','l']]  # SOC(t+1), charging power, upper buffer, lower buffer
            qpform['link']['ineq'] = [[-1, 0, -1], [1, -1, 0]] # -SOC(t) - lowerbuffer < -0.2 and SOC(t) - upperbuffer < 0.8
            qpform['link']['bineq'] = [0, 0]  # Note: the magnitude of the buffer is set later in find_buffer.
        else:
            qpform['states'] = [['x','y','u','l']]  # SOC(t+1), charging power, upper buffer, lower buffer.
            qpform['link']['ineq'] = [[a, -1, 0, 0], [-1, 0, 0, -1], [1, 0, -1, 0]] # -SOC(t) - lowerbuffer < -0.2 qpform.link.ineq = [qp_form.link.ineq, [1, 0, -1, 0]] SOC(t) - upperbuffer < 0.8
            qpform['link']['bineq'] = [0, 0, 0]
        qpform['buffer'] = gen['buffer']
        qpform['u'] = {}
        qpform['l'] = {}
        qpform['u']['lb'] = 0
        qpform['u']['ub'] = 0  # Note: the magnitude of the buffer is set later in find_buffer.
        qpform['u']['h'] = 0
        qpform['u']['f'] = 0
        qpform['l']['lb'] = 0
        qpform['l']['ub'] = 0  # Note: the magnitude of the buffer is set later in find_buffer.
        qpform['l']['h'] = 0
        qpform['l']['f'] = 0
    return qpform


def load_hydro_storage(gen, scale_time):
    """This function loads the parameters for a hydroelectric plant."""
    qpform = {}
    qpform['stor'] = {}
    qpform['output'] = {}
    qpform['x'] = {}
    qpform['y'] = {}
    qpform['link'] = {}
    
    eff = gen['max_gen_capacity'] / (gen['max_gen_flow'] * gen['max_head'] * 84.674) # Power (kW)/ideal power in kW
    qpform['max_gen_flow'] = gen['max_gen_flow']
    qpform['start_wy'] = gen['start_wy_state']
    qpform['stor']['size'] = gen['size_kwh'] * scale_time
    qpform['stor']['self_discharge'] = 0  # Needs to be evaporative losses.
    qpform['stor']['disch_eff'] = 1  # 100% efficient.
    qpform['stor']['usable_size'] = qpform['stor']['size'] * gen['min_head'] / gen['max_head']
    qpform['stor']['power_to_flow'] = 1 / (eff * gen['max_head'] * 84.674)# Power [kW] = efficiency [%] * Flow [1000 ft^3/s] * Head [ft] * 84.674 [kJ/(1000ft^3*ft)]
    qpform['output']['w'] = [[0]]
    qpform['output']['e'] = [[1, 0]]
    qpform['states'] = ['x','y']  # Power and state of charge.
    qpform['x']['lb'] = 0
    qpform['x']['ub'] = gen['max_gen_capacity']
    qpform['x']['h'] = 0
    qpform['x']['f'] = 0

    qpform['y']['lb'] = 0
    qpform['y']['ub'] = qpform['stor']['usable_size']
    qpform['y']['h'] = 0
    qpform['y']['f'] = 0
    qpform['ramp']['b'] = [gen['ramp_down'], gen['ramp_up']]  # Change in power generation.
    qpform['link']['eq'] = [qpform['stor']['power_to_flow'], 0]  # Convert power to flow rate (1000 cfs).
    qpform['link']['b_eq'] = [0]       
    if gen['max_spill_flow'] > 0:
        qpform['states'] = ['x','y','s']  # Add spill flow state.
        qpform['s'] = {}
        qpform['s']['lb'] = 0
        qpform['s']['ub'] = gen['max_spill_flow']
        qpform['s']['h'] = 0
        qpform['s']['f'] = 0
        qpform['link']['eq'] = [[qpform['stor']['power_to_flow'], 0, 1]]
        qpform['link']['b_eq'] = [0]
        qpform['output']['e'] = [[1, 0, 0]]
    if 'buffer' in gen and gen['buffer'] != 0:  # Buffer states.
        if gen['max_spill_flow'] > 0:
            qpform['states'] = ['x','y','s','u','l']
            qpform['link']['eq'] =[[qpform['stor']['power_to_flow'], 0, 1, 0, 0]]
            qpform['link']['ineq'] =[[0, -1, 0, 0, -1], [0, 1, 0, -1, 0]] # -SOC(t)-lowerbuffer < -0.2, SOC(t)-upperbuffer < 0.8
            qpform['output']['e'] = [[1, 0, 0, 0, 0]]
        else:
            qpform['states'] = ['x','y','u','l']
            qpform['link']['eq'] = [[qpform['stor']['power_to_flow'], 0, 0, 0]]
            qpform['link']['ineq'] = [[0, -1, 0, -1], [0, 1, -1, 0]] # -SOC(t)-lowerbuffer < -0.2, SOC(t)-upperbuffer < 0.8
            qpform['output']['e'] = [[1, 0, 0, 0]]
        qpform['link']['bineq'] = [0, 0]

        qpform['u'] = {}
        qpform['u']['lb'] = 0
        qpform['u']['ub'] = 0
        qpform['u']['h'] = 0
        qpform['u']['f'] = 0

        qpform['l'] = {}
        qpform['l']['lb'] = 0
        qpform['l']['ub'] = 0
        qpform['l']['h'] = 0
        qpform['l']['f'] = 0
    return qpform


def load_ac_dc(gen):
    """Load AC/DC generator."""
    qpform = {}
    qpform['output'] = {}
    qpform['a'] = {}
    qpform['b'] = {}
    # First state is AC power transferred to DC power.
    # Second state is DC power transferred to AC power.
    qpform['states'] = [['a','b']]
    qpform['output']['e'] = [[-1, gen['dc_to_ac_eff']]]
    qpform['output']['dc'] = [[gen['ac_to_dc_eff'], -1]]
    qpform['a']['ub'] = gen['capacity']
    qpform['a']['lb'] = 0
    qpform['a']['f'] = 0
    qpform['a']['h'] = 0
    qpform['b']['ub'] = gen['capacity']
    qpform['b']['lb'] = 0
    qpform['b']['f'] = 0
    qpform['b']['h'] = 0
    return qpform


def fit_fcn(outpt, inpt, seg_end, n, order, intercept):
    x_max = [seg_end[0]]
    x_max.extend([seg_end[i+1] - seg_end[i] for i in range(n-1)])
    x = [[0 for j in range(len(outpt))] for i in range(n)]
    for i in range(len(outpt)):
        x[0][i] = min([outpt[i], seg_end[0]])
        for j in range(1,n):
            x[j][i] = max([0, min([x_max[j], outpt[i] - sum([x[k][i] for k in range(j)])])])
    qp = {}
    qp['h'] = [[1 for i in range(order*n)] for j in range(order*n)]
    qp['f'] = [1 for i in range(order*n)]
    for i in range(order*n):
        if i < n:
            a = x[i]
        else:
            a = [x[i-n][j]**2 for j in range(len(x[i-n]))]
        for j in range(i+1):
            if j < n:
                b = x[j]
            else:
                b = [x[j-n][k]**2 for k in range(len(x[j-n]))]
            ab = sum([a[k]*b[k] for k in range(len(a))])
            qp['h'][i][j] = 2*ab
            qp['h'][j][i] = 2*ab
        qp['f'][i] = -2 * sum([a[k]*inpt[k] for k in range(len(a))])

    qp['a'] = [[0 for i in range(order*n)] for j in range(order*n-1)]
    qp['b'] = [0 for j in range(order*n-1)]
    for i in range(n-1):
        qp['a'][i][i] = 1
        qp['a'][i][i+1] = -1
        if order == 2:
            qp['a'][i][n+i] = 2 * x_max[i]
    qp['a_eq'] = [[x_max[i]*1 for i in range(len(x_max))]]
    if order == 2:
        for i in range(n-1, 2*n-1):
            qp['a'][i][i+1] = -1
        qp['a_eq'][0].extend([x_max[i]**2 for i in range(len(x_max))])
        
    qp['b_eq'] = [inpt[-1]]
    qp['lb'] = []
    qp['ub'] = []
    if intercept:
        a = fit_with_intercept(qp,inpt,order,x,n)
    else:
        a,_ = ecos_ls(qp['h'], qp['f'], qp['a'], qp['b'], qp['a_eq'], qp['b_eq'])

    if order == 2:
        for i in range(n, 2*n):
            if a[i] < 1e-6:
                a[i] = 0
    if intercept:
        c_0 = a[0]
        a = a[1:]
    else:
        c_0 = 0

    # plot_fitting(x_max,outpt,inpt,order,A,c_0,n,ones(length(A),1))
    r_square = fit_residual(inpt,outpt,c_0,x,a,order)
    return r_square, a, c_0

def fit_with_intercept(qp,inpt,order,x,n):
    a = [0 for i in range(order*n)]
    for i in range(order*n):
        if i<n:
            a[i] = 2*sum([j for j in x[i]])
        else:
            a[i] = 2*sum([j**2 for j in x[i-n]])
        qp['h'][i].insert(0,a[i])
    a.insert(0,2*len(inpt))
    qp['h'].insert(0,a)
    qp['f'].insert(0,-2*sum(inpt))
    for i in range(order*n-1):
        qp['a'][i].insert(0,0)
    qp['a_eq'][0].insert(0,1)
    a,_ = ecos_ls(qp['h'], qp['f'], qp['a'], qp['b'], qp['a_eq'], qp['b_eq'])
    return a

def fit_residual(inpt,outpt,c_0,x,a,order):
    #find residual of result
    ss_res = 0
    ss_mean = 0
    mean_inpt = sum(inpt)/len(inpt)
    for i in range(len(outpt)):
        if order == 2:
            x2 = [x[j][i] for j in range(len(x))]
            x2.extend([x[j][i]**2 for j in range(len(x))])
            fit_i = c_0 + sum([a[j] for j in range(len(a))])
        else:
            fit_i = c_0 + sum([a[j]*x[j][i] for j in range(len(x))])
        ss_res += (inpt[i] - fit_i)**2
        ss_mean += (inpt[i] - mean_inpt)**2
    r_square = 1 - ss_res / ss_mean
    return r_square

def fit_coproduction(gen, a, b, n):
    """Sets up two least squares problems to find the coefficients of the heat
    co-production."""
    ub = gen['size_kw']
    if 'electricity' in gen['output']:
        lb = gen['startup']['electricity'][-1]
    elif 'direct_current' in gen['output']:
        lb = gen['startup']['direct_current'][-1]
    seg_end = [lb + (ub-lb)*(i+1)/n for i in range(n)]
    c_sort = sorted(range(len(gen['output']['capacity'])), key=lambda k: gen['output']['capacity'][k])
    cap = [gen['output']['capacity'][i] for i in c_sort]
    heat = [gen['output']['heat'][i] for i in c_sort]   
    if 'electricity' in gen['output']:
        elec = [gen['output']['electricity'][i] for i in c_sort]
    elif 'direct_current' in gen['output']:
        elec = [gen['output']['direct_current'][i] for i in c_sort]
    #interpolate input and heat output
    i = 0
    fuel_input = []
    seg_heat = []
    for j in range(len(seg_end)):
        while cap[i+1]<seg_end[j]/ub:
            i+=1
        r = (cap[i+1]-seg_end[j]/ub)/(cap[i+1]-cap[i])
        fuel_input.append(r*seg_end[j]/elec[i] + (1-r)*seg_end[j]/elec[i+1])
        seg_heat.append((r*fuel_input[j]*heat[i] + (1-r)*fuel_input[j]*heat[i+1])*0.95) # Reduce heat co-production by 5% for fitting, because it is better to slightly underestimate.

    cumulative_slope = [seg_heat[i]/seg_end[i] for i in range(len(seg_end))]
    mean_slope = sum(cumulative_slope[:(len(seg_heat)-1)])/(len(seg_heat)-1)
    seg_heat = [min([seg_heat[i],1.3*mean_slope*seg_end[i]]) for i in range(len(seg_end))]
    local_slope = [seg_heat[0]/seg_end[0]]
    for i in range(len(seg_end)-1):
        local_slope.append((seg_heat[i+1]-seg_heat[i])/(seg_end[i+1]-seg_end[i]))

    # Maximum that beta (slope of heat ratio) can increase and remain convex
    # due to convex cost function.
    max_slope_inc = [(a[i+1] - a[i]) / a[i] * local_slope[i] for i in range(len(seg_end)-1)]

    seg_5n = [lb/ub for i in range(n)]
    seg_5n.extend(np.linspace(lb/ub, 1, 5*n+1).tolist())
    seg_fuel_5n = []
    seg_heat_5n = []
    i = 0
    for j in range(6*n+1):
        while cap[i+1]<seg_5n[j]:
            i+=1
        r = (cap[i+1]-seg_5n[j])/(cap[i+1]-cap[i])
        seg_fuel_5n.append(r*seg_5n[j]/elec[i] + (1-r)*seg_5n[j]/elec[i+1])
        seg_heat_5n.append((r*seg_fuel_5n[j]*heat[i] + (1-r)*seg_fuel_5n[j]*heat[i+1])*0.95) 

    heat_out = [[],[]]
    heat_out[0], _ = fit_fcn_2(seg_5n, seg_heat_5n, [seg_end[i]/ub for i in range(len(seg_end))], n, max_slope_inc,None)

    # Repeat for non-zero y-intercept.
    # Slope of 1st segment wih non-zero y-intercept.
    i = 0
    while cap[i+1]<lb/ub:
        i+=1
    r = (cap[i+1]-lb/ub)/(cap[i+1]-cap[i])
    min_fuel = r*lb/elec[i] + (1-r)*lb/elec[i+1]
    min_heat = min_fuel*(r*heat[i] + (1-r)*heat[i+1])

    local_slope[0] = (seg_heat[0]-min_heat)/(seg_end[0]-lb)
    # y_intercept = seg_heat[0] - seg_end[0] * local_slope[0]

    # Maximum that beta (slope of heat ratio) can increase and remain convex
    # due to convex cost function.
    max_slope_inc = [2*(b[i+1] - b[i]) / b[i] * local_slope[i] for i in range(len(seg_end)-1)]
    heat_out[1], h_0 = fit_fcn_2(seg_5n, seg_heat_5n,  [x/ub for x in seg_end], n, max_slope_inc, 1.1*local_slope[0])

    max_heat = max([ub*cap[i]/elec[i]*heat[i] for i in range(len(cap)) if elec[i]>0 and heat[i]>0])
    # max_heat = h_0*ub + seg_end[0]*heat_out[1][0]
    # for i in range(len(heat_out[1])-1):
    #     max_heat += heat_out[1][i+1]*(seg_end[i+1] - seg_end[i])
    return h_0, heat_out, max_heat


def fit_fcn_2(outpt, inpt, seg_end, n, max_slope_inc, local_slope_1):
    """Sets up a least squares problem to find the linear piecewise
    coefficients of the heat co-production."""
    qp = {}
    x_max = [seg_end[0]]
    x_max.extend([seg_end[i+1] - seg_end[i] for i in range(n-1)])
    x = [[0 for j in range(len(outpt))] for i in range(n)]
    for i in range(len(outpt)):
        x[0][i] = min([outpt[i], seg_end[0]])
        for j in range(1,n):
            x[j][i] = max([0, min([x_max[j], outpt[i] - sum([x[k][i] for k in range(j)])])])

    qp['h'] = [[1 for i in range(n)] for j in range(n)]
    qp['f'] = [1 for i in range(n)]
    for i in range(n):
        a = x[i]
        for j in range(i+1):
            b = x[j]
            ab = sum([a[k]*b[k] for k in range(len(a))])
            qp['h'][i][j] = 2*ab
            qp['h'][j][i] = 2*ab
        qp['f'][i] = -2 * sum([a[k]*inpt[k] for k in range(len(a))])

    qp['a'] = [[0 for i in range(n)] for j in range(n-1)]
    qp['b'] = [x*1 for x in max_slope_inc]

    # The sign of this constraint is switched from fit_fcn because we want the
    # coefficients of heat recovery to get smaller.
    for i in range(n-1):
        qp['a'][i][i] = -1
        qp['a'][i][i+1] = 1
    qp['a_eq'] = []
    qp['b_eq'] = []
    if not local_slope_1 is None:
        a = fit_with_intercept2(qp,inpt,local_slope_1,x,n)
        h_0 = a[0]
        fit = a[1:]
    else:
        fit,_ = ecos_ls(qp['h'], qp['f'], qp['a'], qp['b'], qp['a_eq'], qp['b_eq'])
        h_0 = 0
    return fit, h_0

def fit_with_intercept2(qp,inpt,local_slope,x,n):
    a = [0] * n
    for i in range(n):
        a[i] = 2*sum([j for j in x[i]])
        qp['h'][i].insert(0,a[i])
        
    lsc = [0,1]
    for i in range(n-1):
        qp['a'][i].insert(0,0)
        lsc.append(0)
    a.insert(0,2*len(inpt))
    qp['h'].insert(0,a)
    qp['f'].insert(0,-2*sum(inpt))
    qp['a'].append(lsc)
    qp['b'].append(local_slope)
    a,_ = ecos_ls(qp['h'], qp['f'], qp['a'], qp['b'], qp['a_eq'], qp['b_eq'])
    return a

def max_ac_dc_gen(subnet,qpform):
    #identify upper bound for utility states (helps with scaling)
    max_dc_gen = 0
    gen_names = [qpform[i]['name'] for i in range(len(qpform))]
    if 'direct_current' in subnet:
        nn = len(subnet['direct_current']['nodes'])
        for n in range(nn):
            equip = subnet['direct_current']['equipment'][n]
            ne = len(equip)
            for j in range(ne):
                k = gen_names.index(equip[j])
                if qpform[k]['type'] == 'HydroStorage':
                    max_dc_gen += qpform[k]['x']['ub']
                elif qpform[k]['type'] == 'ElectricStorage':
                    max_dc_gen += qpform[k]['stor']['peak_disch']
                elif not qpform[k]['type'] == 'ACDCConverter':
                    if qpform[k]['type'] == 'Solar': #Solar
                        max_dc_gen += qpform[k]['size_kw']*qpform[k]['output']['dc'][0][0]
                    else:
                        if len(qpform[k]['output']['dc'][-1]) == 1:
                            dc_output = [qpform[k]['output']['dc'][-1][0] for s in range(len(qpform[k]['states'][-1]))]
                        else:
                            dc_output = qpform[k]['output']['dc'][-1]
                        s_ub = [qpform[k][s]['ub'][-1] for s in qpform[k]['states'][-1]]
                        max_dc_gen += max([0,sum([s_ub[i]*dc_output[i] for i in range(len(s_ub))])])
    max_ac_gen = 0
    if 'electrical' in subnet:
        nn = len(subnet['electrical']['nodes'])
        for n in range(nn):
            equip = subnet['electrical']['equipment'][n]
            ne = len(equip)
            for j in range(ne):
                k = gen_names.index(equip[j])
                if qpform[k]['type'] == 'HydroStorage':
                    max_ac_gen += qpform[k]['x']['ub']
                elif qpform[k]['type'] == 'ElectricStorage':
                    max_ac_gen += qpform[k]['stor']['peak_disch']
                elif not qpform[k]['type'] == 'ACDCConverter':
                    if len(qpform[k]['states']) == 0: #Solar
                        max_ac_gen += qpform[k]['size_kw']*qpform[k]['output']['e'][0][0]
                    else:
                        if len(qpform[k]['output']['e'][-1]) == 1:
                            ac_output = [qpform[k]['output']['e'][-1][0] for s in range(len(qpform[k]['states'][-1]))]
                        else:
                            ac_output = qpform[k]['output']['e'][-1]
                        s_ub = []
                        for s in qpform[k]['states'][-1]:
                            if isinstance(qpform[k][s]['ub'],list):
                                s_ub.append(qpform[k][s]['ub'][-1])
                            else:
                                s_ub.append(qpform[k][s]['ub'])
                        max_ac_gen += max([0,sum([s_ub[i]*ac_output[i] for i in range(len(s_ub))])])
    return max_ac_gen, max_dc_gen

def update_utility_sellback_limit(qpform,test_data,max_ac_gen, max_dc_gen):
    n_g = len(qpform)
    for i in range(n_g):
        if qpform[i]['type'] == 'Utility'and len(qpform[i]['states'][0])>0: #avoid things like gas utility with no states
            #identify the network
            out = [*qpform[i]['output']] 
            if 'e' in out:
                net = 'electrical'
            elif 'h' in out:
                net = 'district_heat'
            elif 'c' in out:
                net = 'district_cool'
            elif 'w' in out:
                net = 'hydro'
            d = []
            if net in test_data['nodedata_network_info']:
                for n in test_data['nodedata_network_info'][net]:
                    if 'demand_history' in n:
                        #To accomodate different time stamps on history it will just add the maximum of each node, even if maximums are not coeincident.
                        d.append(float(max(test_data['nodedata'].read(n['demand_history'][0],0,-1, field = n['demand_history'][1]))))
            if len(d)>0:
                qpform[i]['x']['ub'] = 10*sum(d) #max Purchase
            else:
                qpform[i]['x']['ub'] = 1e6 #arbitrary upper bound that is not inf
            qpform[i]['ub'] = qpform[i]['x']['ub']
            if net == 'electrical' and 'y' in qpform[i]:
                qpform[i]['y']['ub'] = max_ac_gen + max_dc_gen
        if qpform[i]['type']=='ACDCConverter':
            qpform[i]['a']['ub'] = max_ac_gen
            qpform[i]['b']['ub'] = max_dc_gen
    return qpform

def load_solar(gen,subnet):
    # There are no states or outputs for solar because renewable
    # outputs are handled on the demand side.
    qpform = {}
    qpform['azimuth'] = gen['azimuth']
    qpform['eff'] = gen['eff']
    qpform['pv_type'] = gen['pv_type']
    qpform['size_m2'] = gen['size_m2']
    qpform['size_kw'] = gen['size_kw']
    qpform['tilt'] = gen['tilt']
    qpform['tracking'] = gen['tracking']
    qpform['states'] = [[]]
    qpform['output'] = {}
    if 'direct_current' in subnet and any(gen['name'] in agg
            for agg in subnet['direct_current']['equipment']):
        qpform['output']['dc'] = [[1]]
    else:
        qpform['output']['e'] = [[1]]
    qpform['location'] = {}
    
    if 'direct_current' in subnet:
        for n in range(len(subnet['direct_current']['nodes'])):
            if gen['name'] in subnet['direct_current']['equipment'][n]:
                qpform['location'] = subnet['direct_current']['location'][n]
                break
    if not 'latitude' in qpform['location']:
        for n in range(len(subnet['electrical']['nodes'])):
            if gen['name'] in subnet['electrical']['equipment'][n]:
                qpform['location'] = subnet['electrical']['location'][n]
                break
    return qpform


import datetime
from eagers.solver.sort_solution import soc_2_power

def update_observer(gen,subnet,observer,date,options,solution,actual_data,forecast):   
    trim_history(gen,observer['history'],date[0],24)
    observer_short_history(gen,date,actual_data,subnet,solution,observer['history'])
    if options['method'] == 'planning': #assumes forecast is perfect
        n_s = len(date)-1
        date = [date[t] + datetime.timedelta(hours=options['horizon']) for t in range(len(date))] #count forward by len of the horizon, rounded to nearest second
    elif options['method'] in ['dispatch', 'control']:
        observer_future(gen,date,subnet,solution,forecast,observer['future'])
        n_s = 1 
        date =  [date[t] + datetime.timedelta(hours=options['resolution']) for t in range(len(date))] #count forward 1 step, rounded to nearest second
    observer['timestamp'] = date[0]
    n_g = len(gen)
    for i in range(n_g):
        if gen[i]['type'] in ['ElectricGenerator','CombinedHeatPower','Chiller','Heater','Electolyzer','HydrogenGenerator','CoolingTower']:
            observer['gen_state'][i] = solution['dispatch'][i][n_s]
            observer['gen_status'][i] = solution['dispatch'][i][n_s]>0
        elif gen[i]['type'] in ['ElectricStorage','ThermalStorage','HydrogenStorage']:
            observer['stor_state'][i] = solution['dispatch'][i][n_s]
        elif gen[i]['type'] in ['HydroStorage']:
            n = gen[i]['subnet_node']['hydro'] #dam #
            observer['stor_state'][i] = solution['hydro_soc'][n][n_s-1]  
    for i,k in enumerate(solution['line_flow']):
        observer['line_flow'][i] = solution['line_flow'][k][n_s-1]
    for i,k in enumerate(solution['fluid_loop']):
        observer['fluid_loop_temperature'][i] = solution['fluid_loop'][k][n_s-1]
    return date

def observer_short_history(gen,date,actual_data,subnet,solution,history):
    dt = [(date[t+1] - date[t]).total_seconds()/3600 for t in range(len(date)-1)]
    n_s = len(actual_data['timestamp'])
    history['timestamp'].extend(date[1:n_s+1])
    for i,k in enumerate(history['generator_state']):
        if 'stor' in gen[i]:
            sd = soc_2_power(gen[i]['stor'],solution['dispatch'][i],dt)
            history['generator_state'][k].extend(sd[:n_s])
            history['storage_state'][k].extend(solution['dispatch'][i][1:n_s+1])
        else:
            history['generator_state'][k].extend(solution['dispatch'][i][1:n_s+1])
    for k in history['line_flow']:
            history['line_flow'][k].extend(solution['line_flow'][k][:n_s])
    for k in history['fluid_loop']:
            history['fluid_loop'][k].extend(solution['fluid_loop'][k][:n_s])
    for net in subnet['network_names']:
        for an in subnet[net]['nodes']:
            for n in an:
                if n in actual_data:
                    headers = list(actual_data[n].keys())
                    headers.remove('network')
                    for h in headers:
                        if not h in history['nodedata'][n]:
                            history['nodedata'][n][h] = []
                        history['nodedata'][n][h].extend(actual_data[n][h][:n_s])
    for k in list(actual_data['weather'].keys()):
        history['weather'][k].extend(actual_data['weather'][k][:n_s])
    if 'lb_relax' in solution:
        if 'lb_relax' not in history:
            history['lb_relax'] = []
        history['lb_relax'].extend(solution['lb_relax'][:n_s])

def observer_future(gen,date,subnet,solution,forecast,future):
    dt = [(date[t+1] - date[t]).total_seconds()/3600 for t in range(len(date)-1)]
    future['timestamp'] = forecast['timestamp'][1:]
    for i,k in enumerate(future['generator_state']):
        if k in future['storage_state']:
            sd = soc_2_power(gen[i]['stor'],solution['dispatch'][i],dt)
            future['generator_state'][k] = sd[1:]
            future['storage_state'][k] = solution['dispatch'][i][2:]
        else:
            future['generator_state'][k] = solution['dispatch'][i][2:]
    for k in future['line_flow']:
        future['line_flow'][k] = solution['line_flow'][k][1:]
    for k in future['fluid_loop']:
        future['fluid_loop'][k] = solution['fluid_loop'][k][1:]
    for net in subnet['network_names']:
        for an in subnet[net]['nodes']:
            for n in an:
                if n in forecast:
                    headers = list(forecast[n].keys())
                    headers.remove('network')
                    for h in headers:
                        if not h in future['nodedata'][n]:
                            future['nodedata'][n][h] = []
                        future['nodedata'][n][h].append(forecast[n][h][1:])
    for k in list(forecast['weather'].keys()):
        future['weather'][k] = forecast['weather'][k][1:]
    if 'lb_relax' in solution:
        future['lb_relax'] = solution['lb_relax'][1:]


def trim_history(gen,history,date,length):
    h_time = history['timestamp']
    td = 0
    while h_time[td]<(date-datetime.timedelta(hours=length)):
        td +=1
    if td>0:
        headers = list(history.keys())
        for h in headers:
            if isinstance(history[h],list) and len(history[h])>0:
                if isinstance(history[h][0],(float,int)) or h == 'timestamp':
                    del history[h][0:td]
                else:
                    for x in range(len(history[h])):
                        del history[h][x][0:td]
            elif isinstance(history[h],dict):
                sub_headers = list(history[h].keys())
                for sh in sub_headers:
                    if isinstance(history[h][sh],list) and len(history[h][sh])>0:
                        if isinstance(history[h][sh][0],(float,int)):
                            del history[h][sh][0:td]
                        else:
                            for x in range(len(history[h][sh])):
                                del history[h][sh][x][0:td]


"""Logic for calculating net cost.
"""

from eagers.update.update_cost import update_cost


def net_cost(gen, dispatch, date, market):
    """Net cost.

    Positional arguments:
    gen - (list of dict) QP form of each component.
    dispatch - (list of list) Dispatch solution for each component.
    date - (list of datetime) Timestamps to calculate cost for.
    market - Market data.
    """
    n_g = len(gen)
    gen_cost = [[0] for i in range(n_g)]
    scale_cost = update_cost(date[1:], gen, market)
    n_s = len(date) - 1
    dt = [(date[t+1] - date[t]).seconds/3600 for t in range(n_s)]
    for i, g in enumerate(gen):
        if g['type'] in ['ElectricGenerator','CombinedHeatPower','Heater', 'CoolingTower', 'Electrolyzer', 'HydrogenGenerator']:                
            states = g['states'][-1]
            for t in range(n_s):
                p = 0
                j = 0
                c = 0
                while p < dispatch[i][t+1] and j < len(states):
                    x = min([g[states[j]]['ub'][-1],dispatch[i][t+1]-p])
                    p +=x
                    c += (x*g[states[j]]['f'][-1] + x*x*g[states[j]]['h'][-1])*scale_cost[i][t]*dt[t]
                    j +=1
                if 'start_cost' in g and dispatch[i][t+1]>0 and dispatch[i][t]==0:
                    c += g['start_cost']
                gen_cost[i].append(c)
        elif g['type'] == 'Utility':
            for t in range(n_s):
                if dispatch[i][t] >= 0:
                    c = dispatch[i][t] * scale_cost[i][t] * dt[t]
                elif 'sellback_rate' in g:
                    if g['sellback_rate'] > 0:
                        # Constant sellback rate.
                        c = dispatch[i][t] * g['sellback_rate'] * dt[t]
                    else:
                        c = dispatch[i][t] * g['sellback_perc'] / 100 * scale_cost[i][t] * dt[t]
                else:
                    c = 0
                gen_cost[i].append(c)
        elif g['type'] == 'Tradepoint':
            for t in range(n_s):
                six_param = scale_cost[i][t]
                x = dispatch[i][t]
                if x>0:
                    c = (x*six_param[0] + x*x*six_param[1])*dt[t]
                elif x<0:
                    c = (x*six_param[3] + x*x*six_param[4])*dt[t]
                gen_cost[i].append(c)
    cost = sum([sum(gen_cost[i]) for i in range(n_g)])
    return cost



import numpy as np
from scipy import sparse
from scipy.linalg import sqrtm
import ecos
# Notes on speed https://quantsrus.github.io/post/state_of_convex_quadratic_programming_solvers/


def ecos_qp(H, f, A, b, Aeq, beq, ub, lb, x_keep, r_keep, req_keep):
    ''' 
    Refer to details in thesis: https://www.research-collection.ethz.ch/handle/20.500.11850/74559

    Need to convert to second order conic problem following template for matlab ecos:
    https://github.com/embotech/ecos-matlab/blob/master/bin/ecosqp.m
    See Linear Cone Program example on https://cvxopt.readthedocs.io/en/latest/coneprog.html for creation of G and h
    Other info:
    https://math.stackexchange.com/questions/1932532/converting-from-quadratic-to-second-order-cone-optimization-problem
    https://inst.eecs.berkeley.edu/~ee227a/fa10/login/l_conic_socp.html 
    
    so the states x are expanded with y states representing xHx anytime H[i]>0 so the translation involves:
    x = [x;y] 
    c = [f; 1]
    Aeq = [Aeq,zeros(x)]
    beq = beq
    U = sqrt(H)  ** easy if H is diagonal and all values are positive
    G = [[A , zeros(x)]             h = [[b;]
        [-eye(x); zeros(y)]             [-lb;]
        [eye(x); zeros(y)]              [ub;]
        [0, -1/sqrt(2)]                 [1/sqrt(2);]
        [-U; zeros(x)]                  [zeros(x);]
        [0; -1/sqrt(2)]                 [1/sqrt(2);]] (second order)
    '''
    x_scale, eq_scale, ineq_scale = scale_factors(H, f, A, b, Aeq, beq, ub, lb)
    x_l = sum(x_keep)# number of states
    s_num = [None for i in range(len(x_keep))]
    s_rev = []
    s=0
    for i in range(len(x_keep)):
        if x_keep[i]:
            s_num[i] = s
            s_rev.append(i)
            s+=1
    #create triplets for A and A_eq
    val_Aeq = []
    row_Aeq = []
    col_Aeq = []
    req=0
    for row in range(len(Aeq)):
        if req_keep[row]:
            for i in range(len(Aeq[row])):
                if Aeq[row][i]!=0 and not s_num[i] is None:
                    val_Aeq.append(Aeq[row][i]*x_scale[i]/eq_scale[row])
                    row_Aeq.append(req)
                    col_Aeq.append(s_num[i])
            req += 1
    

    val_G = []
    row_G = []
    col_G = []
    r=0
    for row in range(len(A)):
        if r_keep[row]:
            for i in range(len(A[row])):
                if A[row][i]!=0 and not s_num[i] is None:
                    val_G.append(A[row][i]*x_scale[i]/ineq_scale[row])
                    row_G.append(r)
                    col_G.append(s_num[i])
            r+= 1

    if not lb is None:
        r = upper_lower_bound(val_G,row_G,col_G,r,x_l)
    f = [f[i]*x_scale[i] for i in range(len(f))]
    c_scale = max([max(f),-min(f)])
    f = [ff/c_scale for ff in f]        
    if isinstance(H[0],list):
        H = [[H[i][j]*x_scale[j]**2/c_scale for j in range(len(H[i]))] for i in range(len(H))]
        cones = quadratic_cost_constraint(H,val_G,row_G,col_G,s_num,x_keep,r)
    else:
        H = [H[i]*x_scale[i]**2/c_scale for i in range(len(H))]    
        cones = diag_quadratic_cost_constraint_alt(H,val_G,row_G,col_G,s_num,x_keep,r)
    ineq_rows = r+sum([c+2 for c in cones])
    states = x_l + len(cones)
    G = sparse.csc_matrix((np.array(val_G),(np.array(row_G),np.array(col_G))), shape = (ineq_rows,states),dtype='float64')

    h = ineq(b,ub,lb,r_keep,x_keep,cones,x_scale,ineq_scale)

    dims = {}
    dims['l'] = r
    dims['q'] = [c+2 for c in cones]
    dims['e'] = 0

    c = [ff for i,ff in enumerate(f) if x_keep[i]]
    c.extend([1 for i in range(len(cones))])
    c = np.array(c)

    A_ecos = sparse.csc_matrix((np.array(val_Aeq),(np.array(row_Aeq),np.array(col_Aeq))), shape = (req,states),dtype='float64')
    b_ecos = np.array([beq[i]/eq_scale[i] for i in range(len(beq)) if req_keep[i]])
    
    solution = ecos.solve(c, G, h, dims, A_ecos, b_ecos, verbose=False)

    # Exit Flags:
    # 0: optimal
    # 1: primal infeasible
    # 2: dual infeasible
    # 10: close to optimal solution found
    # -1: maximum number of iterations reached
    # -2: numerical problems (unreliable search direction)
    # -3: numerical problems (slacks or multipliers became exterior)
    # -7: unknown problem in solver

    if solution['info']['exitFlag']==0 or solution['info']['exitFlag'] == 10:
        feasible = 0
    else:
        feasible = solution['info']
        # feasible = solution['info']['exitFlag']
    #add removed states back in as zero
    x = [0 for i in range(len(x_keep))]
    for i in range(x_l):
        x[s_rev[i]] = float(solution['x'][i]*x_scale[s_rev[i]])
    return x, feasible


def quadratic_cost_constraint(H,val_G,row_G,col_G,s_num,x_keep,r):
    x_l = sum(x_keep)# number of states
    
    H_ar = np.array(H)
    U_ar = sqrtm(H_ar)
    U = [[float(U_ar[i,j]) for i in range(x_l)] for j in range(x_l)]
    
    val_G.append(-1/np.sqrt(2))
    row_G.append(r)
    col_G.append(x_l)

    #create a single cone for all of the quadratic terms
    cones = [0]
    for row in range(len(U)):
        if x_keep[row]:
            for i in range(len(U[row])):
                if x_keep[i] and U[row][i]!=0:         
                    #add the state constraint (2nd order part on LHS of inequality)
                    val_G.append(-U[row][i])
                    row_G.append(r+cones[0]+1)
                    col_G.append(s_num[i])
                if any([j!=0 for j in U[row]]):
                    cones[0] += 1

    val_G.append(1/np.sqrt(2))
    row_G.append(r + cones[0] + 1)
    col_G.append(x_l)
    return cones


def diag_quadratic_cost_constraint(H,val_G,row_G,col_G,s_num,x_keep,r):
    x_l = sum(x_keep)# number of states
    # add the first cone constraint     
    val_G.append(-1/np.sqrt(2))
    row_G.append(r)
    col_G.append(x_l)

    #create a single cone for all of the quadratic terms
    cones = [0]
    for i in range(len(x_keep)):
        if x_keep[i] and H[i]>0:          
            #add the state constraint (2nd order part on LHS of inequality)
            val_G.append(-float(np.sqrt(H[i])))
            row_G.append(r+cones[0]+1)
            col_G.append(s_num[i])
            cones[0] += 1

    val_G.append(1/np.sqrt(2))
    row_G.append(r + cones[0] + 1)
    col_G.append(x_l)
    
    return cones


def upper_lower_bound(val_G,row_G,col_G,r,x_l):
    #add upper and lower bounds
    val_G.extend([-1 for i in range(x_l)])
    row_G.extend([r+i for i in range(x_l)])
    col_G.extend([i for i in range(x_l)])
    val_G.extend([1 for i in range(x_l)])
    row_G.extend([r+x_l+i for i in range(x_l)])
    col_G.extend([i for i in range(x_l)])
    r += 2*x_l
    return r


def ineq(b,ub,lb,r_keep,x_keep,cones,x_scale,ineq_scale):
    h = [b[r]/ineq_scale[r] for r in range(len(b)) if r_keep[r]]
    if not lb is None:
        h.extend([-lb[j]/x_scale[j] for j in range(len(lb)) if x_keep[j]])
        h.extend([ub[j]/x_scale[j] for j in range(len(lb)) if x_keep[j]])
    for j in range(len(cones)):
        h.append(1/np.sqrt(2))
        h.extend([0 for j in range(cones[j])])
        h.append(1/np.sqrt(2))
       
    h = np.array(h)
    return h


def diag_quadratic_cost_constraint_alt(H,val_G,row_G,col_G,s_num,x_keep,r):
    '''
    Tried making smaller cones (faster?)
    '''
    states = sum(x_keep)# number of states
    cones = [1 for i in range(len(x_keep)) if x_keep[i] and H[i]>0]
    #create a new cone for each nonzero quadratic cost term
    for i in range(len(x_keep)):
        if x_keep[i] and H[i]>0:  
            val_G.append(-1/np.sqrt(2))
            row_G.append(r)
            col_G.append(states)

            #add the state constraint (2nd order part on LHS of inequality)
            val_G.append(-float(np.sqrt(H[i])))
            row_G.append(r + 1)
            col_G.append(s_num[i])

            val_G.append(1/np.sqrt(2))
            row_G.append(r + 2)
            col_G.append(states)

            states +=1
            r +=3
    return cones


def ecos_scaling(H, f, A, b, Aeq, beq, ub, lb):
    xl = len(f)
    if not ub is None:
        x_scale = [max([abs(ub[i]), abs(lb[i])]) for i in range(xl)]
        x_scale = [s if s!=0 else 1 for s in x_scale]
    else:
        x_scale = [1 for i in range(xl)]
    H2 = [H[i]*x_scale[i]**2 for i in range(xl)]
    f2 = [f[i]*x_scale[i] for i in range(xl)]
    ub2 = [ub[i]/x_scale[i] if ub[i]>0 else 1e-6 for i in range(xl)]
    lb2 = [lb[i]/x_scale[i] for i in range(xl)]
    A2 = [[row[i]*x_scale[i] for i in range(xl)] for row in A]
    Aeq2 = [[row[i]*x_scale[i] for i in range(xl)] for row in Aeq]
    b2 = [j for j in b]
    beq2 = [j for j in beq]
    for r,row in enumerate(A2):
        row_nz = [abs(j) for j in row if j!=0]
        if abs(b[r])>0:
            row_nz.append(abs(b[r]))
        r_scale = float(np.sqrt(min(row_nz)*max(row_nz)))
        if r_scale !=1:
            A2[r] = [j/r_scale for j in row]
            b2[r] = b2[r]/r_scale
    for r,row in enumerate(Aeq2):
        row_nz = [abs(j) for j in row if j!=0]
        if abs(beq[r])>0:
            row_nz.append(abs(beq[r]))
        r_scale = float(np.sqrt(min(row_nz)*max(row_nz)))
        if r_scale !=1:
            Aeq2[r] = [j/r_scale for j in row]
            beq2[r] = beq2[r]/r_scale
    return H2, f2, A2, b2, Aeq2, beq2, ub2, lb2, x_scale


def scale_factors(H, f, A, b, Aeq, beq, ub, lb):
    xl = len(f)
    if not ub is None:
        x_scale = [max([abs(ub[i]), abs(lb[i])]) for i in range(xl)]
        x_scale = [s if s!=0 else 1 for s in x_scale]
    else:
        x_scale = [1 for i in range(xl)]
    eq_scale = [1 for j in beq]
    ineq_scale = [1 for j in b]
    for r,row in enumerate(A):
        row_nz = [abs(j*x_scale[i]) for i,j in enumerate(row) if j!=0]
        if abs(b[r])>0:
            row_nz.append(abs(b[r]))
        ineq_scale[r] = float(np.sqrt(min(row_nz)*max(row_nz)))
    for r,row in enumerate(Aeq):
        row_nz = [abs(j*x_scale[i]) for i,j in enumerate(row) if j!=0]
        if abs(beq[r])>0:
            row_nz.append(abs(beq[r]))
        eq_scale[r] = float(np.sqrt(min(row_nz)*max(row_nz)))
    return x_scale, eq_scale, ineq_scale



"""Configuration for the Excel interface."""

from openpyxl.styles import Font, NamedStyle, PatternFill


# File suffixes.
# Cf. https://www.quora.com/What-is-the-extension-of-Excel-files
WORKBOOK_SUFFIX = '.xlsx'

# Values.
# Optimoptions.
AUTO_STARTDATE_VALUES = ('auto', 'automatic')

# Help tab.
HELP_SECTIONHEADER_WIDTH = 8  # cells
HELP_TAB_COLOR = 'de781f'

# NamedStyles.
# header
header = NamedStyle(name='header')
header.font = Font(color='ffffff', bold=True)
header.fill = PatternFill(fill_type='solid', start_color='a0a0a0')
# help_sectionheader
help_sectionheader = NamedStyle(name='help_sectionheader')
help_sectionheader.font = Font(bold=True)
help_sectionheader.fill = PatternFill(fill_type='solid', start_color='dfebe4')

STYLES = (header, help_sectionheader)



'''
Loads equipment sizes as determined by EnergyPlus
'''

from eagers.config.building import EIO_FIELDS


def import_sizes(filename,building):
    component = parse_eio(filename)
    for t in range(len(component['type'])):
        i_loc = EIO_FIELDS[component['type'][t]]['location']
        if len(i_loc)==2:
            if i_loc[1] in getattr(building,i_loc[0]):
                i = getattr(building,i_loc[0])[i_loc[1]].index(component['name'][t])
            else:
                i = component['name'][t]
        elif len(i_loc)==3:
            i = getattr(building,i_loc[0])[i_loc[1]][i_loc[2]].index(component['name'][t])        
        
        try:
            par = EIO_FIELDS[component['type'][t]][component['field'][t]][0]
            if len(i_loc)==1:
                getattr(building,i_loc[0])[component['name'][t]][par] = component['value'][t]
            elif len(i_loc)==2:
                if i_loc[1]=='name' and 'name' not in getattr(building,i_loc[0]):
                    getattr(building,i_loc[0])[i][par] = component['value'][t]
                else:
                    getattr(building,i_loc[0])[par][i] = component['value'][t]
            elif len(i_loc)==3:
                getattr(building,i_loc[0])[i_loc[1]][par][i] = component['value'][t]
        except KeyError:
            if component['field'][t]=='Design Size Reheat Coil Sizing Air Volume Flow Rate [m3/s]':
                par = 'rated_air_flow'
            elif component['field'][t]=='Design Size Reheat Coil Sizing Inlet Air Temperature [C]':
                par = 'air_inlet_temperature'
            elif component['field'][t]=='Design Size Reheat Coil Sizing Inlet Air Humidity Ratio [kgWater/kgDryAir]':
                par = 'air_inlet_humidity_ratio'
            building.coils_heating[building.hvac['terminals']['reheat_coil_name'][i]][par]= component['value'][t]
        
    cp_air = 1025 # J/kg*K
    air_density = 1.204 #kg/m^3 %standard air density (1.204 kg/m3) adjusted for the i_local barometric pressure (standard barometric pressure corrected for altitude, ASHRAE 1997 HOF pg. 6.1).
    for c in building.coils_heating:
        if building.coils_heating[c]['type'] == 'Heating:Water' and building.coils_heating[c]['rated_air_flow'] == 0:
            building.coils_heating[c]['rated_air_flow'] = building.coils_heating[c]['design_capacity']/(cp_air*air_density*(building.coils_heating[c]['air_outlet_temperature']-building.coils_heating[c]['air_inlet_temperature']))

    for c in building.coils_cooling:
        if building.coils_cooling[c]['type'] == 'Cooling:Water' and building.coils_cooling[c]['rated_air_flow'] == 0:
            building.coils_cooling[c]['rated_air_flow'] = building.coils_cooling[c]['design_capacity']/(cp_air*air_density*(building.coils_cooling[c]['air_inlet_temperature']-building.coils_cooling[c]['air_outlet_temperature']))

    building.coils_cooling = water_setpoint(building.coils_cooling,'c', building.plant_demand_equip, building.plant_supply_equip,building.plant_loop)
    building.coils_heating = water_setpoint(building.coils_heating,'h', building.plant_demand_equip, building.plant_supply_equip,building.plant_loop)
    return building


def parse_eio(filename):
    heading,z_info = read_eio(filename)
    component = {}
    c_objects = z_info['Component Sizing Information']
    var = ['type','name','field','value']
    for i in var:
        component[i] = []
    for j in range(len(c_objects)):
        component['type'].append(read_num(c_objects[j][0]))
        component['name'].append(read_num(c_objects[j][1].lower()))
        component['field'].append(read_num(c_objects[j][2]))
        component['value'].append(read_num(c_objects[j][3]))
        
    if 'Water Heating Coil Capacity Information' in z_info:
        hc_objects = z_info['Water Heating Coil Capacity Information']
        for j in range(len(hc_objects)):
            component['type'].append(hc_objects[j][0])
            component['name'].append(hc_objects[j][1].lower())
            component['field'].append(heading['Water Heating Coil Capacity Information'][1])
            component['value'].append(read_num(hc_objects[j][2]))
    if 'Water Cooling Coil Capacity Information' in z_info:
        cc_objects = z_info['Water Cooling Coil Capacity Information']
        coil_cool = heading['Water Cooling Coil Capacity Information']
        f = len(coil_cool)-1
        for p in range(len(cc_objects)):
            component['type'].extend([cc_objects[p][0] for x in range(f)])
            component['name'].extend([cc_objects[p][1].lower() for x in range(f)])
            component['field'].extend([i for i in coil_cool[1:]])
            component['value'].extend([read_num(cc_objects[p][x+2]) for x in range(f)])
    return component


def read_eio(filename):
    """Parses the given text file into a dictionary with the relevant
    data.
    """
    # Read lines of file.
    heading = {}
    z_info = {}
    with open(filename, 'r') as f:
        line = next(f)
        while True: #START with second line
            try:
                line = next(f).lstrip()
            except StopIteration:
                break
            com = [pos for pos, char in enumerate(line) if char == ',']
            if line[0]=='!':
                cat = line[3:com[0]-1]
                cat = cat.replace('<','')
                cat = cat.replace('#','')
                if len(com)==1:
                    headings = [line[com[0]+2:-1].lstrip()]
                else:
                    headings = [line[com[0]+1:com[1]].lstrip()]
                    for j in range(2,len(com)-1):
                        headings.append(line[com[j]+1:com[j+1]].lstrip())
                    headings.append(line[com[-1]+1:-1].lstrip())
                heading[cat] = headings
                z_info[cat]= []
            elif len(com)>0:
                cat = line[:com[0]]
                cat = cat.replace('<','')
                cat = cat.replace('#','')
                if len(com)==1:
                    info=[line[com[0]+1:-1].lstrip()]
                else:
                    info=[line[com[0]+1:com[1]].lstrip()]
                    for j in range(1,len(com)-1):
                        info.append(line[com[j]+1:com[j+1]].lstrip())
                    info.append(line[com[-1]+1:-1].lstrip())
                try:
                    z_info[cat].append(info)
                except KeyError:
                        pass#hopefully none of these are ones I need
    return heading,z_info


def water_setpoint(coil,c_or_h,de,se,bpl):
    for x in coil:
        if coil[x]['type'] == 'Cooling:Water:DetailedGeometry' or coil[x]['type']== 'Cooling:Water':
            try:
                pl = de['loop'][de['name'].index(x)]
            except ValueError:
                pl = se['loop'][se['name'].index(x)]
            if 'water_inlet_temperature' in coil[x]: 
                Tw_in = coil[x]['water_inlet_temperature']
            else:
                Tw_in = bpl['exit_temperature'][pl]
                coil[x]['water_inlet_temperature'] = Tw_in
            if c_or_h == 'c':
                Tw_out = Tw_in + bpl['temperature_difference'][pl]
            elif c_or_h == 'h':
                Tw_out = Tw_in - bpl['temperature_difference'][pl]
            coil[x]['water_outlet_temperature'] = Tw_out
    return coil

def read_num(n):
    """Convert to an integer or float if possible."""
    try:
        n = int(n)
    except ValueError:
        try:
            n = float(n)
        except ValueError:
            if isinstance(n, str):
                if n:
                    return n
                else:
                    return None
    return n


from eagers.building.components.water_equipment import water_equipment

def flow_resolver_demand(building,node,profile,l,T_mains):
    ## resolve loop flows
    #see details starting on pg 461
    d_equip = building.plant_demand_equip 
    on_loop = [i for i,x in enumerate(d_equip['loop']) if x == l]
    d_pump = None
    ## flow requests of water equipment first so upstream splitter is correct
    for j in on_loop:
        if d_equip['type'][j]==  'wateruse:connections':
            in_n = d_equip['inlet'][j]
            out_n = d_equip['outlet'][j]
            flow,_,_,_,_ = water_equipment(building.water_use[d_equip['name'][j]],profile,T_mains,node['demand_temperature'][in_n]) 
            z = building.zones['name'].index(building.water_use[d_equip['name'][j]]['zone'])
            node['demand_flow'][out_n] = flow*building.zones['multiplier'][z]
            node['demand_flow'][in_n] = node['demand_flow'][out_n] 
        elif d_equip['type'][j][:5]== 'coil:':
            in_n = d_equip['inlet'][j]
            out_n = d_equip['outlet'][j]
            node['demand_flow'][in_n] = node['demand_flow'][out_n] 

    ##connect flow requests from individual components
    for j in on_loop:
        in_n = d_equip['inlet'][j]
        out_n = d_equip['outlet'][j]
        if d_equip['type'][j]== 'mixer':
            if type(in_n) != list:
                node['demand_flow'][out_n] = node['demand_flow'][in_n]
            else:
                node['demand_flow'][out_n] = sum([node['demand_flow'][k] for k in in_n]) 
        elif d_equip['type'][j]== 'splitter':
            if type(out_n) != list:
                node['demand_flow'][in_n] = node['demand_flow'][out_n]
            else:
                node['demand_flow'][in_n] = sum([node['demand_flow'][k] for k in out_n]) 
        elif d_equip['type'][j][:5]== 'pump:':
            d_pump = j
        elif d_equip['type'][j][:5]== 'pipe:': 
            node['demand_flow'][out_n] = max(node['demand_flow'][in_n],node['demand_flow'][out_n]) #branch flow determined by component with largest request
            node['demand_flow'][in_n] = node['demand_flow'][out_n] 
        elif not(d_equip['type'][j][:5]== 'coil:' or d_equip['type'][j] == 'wateruse:connections'):
            node['demand_flow'][in_n] = node['demand_flow'][out_n] 
            # print(d_equip['type'][j])
        #flow is already set by component (coil or water equipment)
    loop_flow_request = node['demand_flow'][building.plant_demand_nodes['outlet_node'][l]]/building.plant_loop['fluid_density'][l] # convert mass flow to volumetric flow (parameters saved in building file in volumetric flow)

    #find pump & supply limits
    pump_flow = 0 
    demand_bypass_flow = 0 
    if d_pump==None:
        p = [building.plant_supply_equip['name'][i] for i,x in enumerate(building.plant_supply_equip['type']) if x[:5]=='pump:' and building.plant_supply_equip['loop'][i] == l]
        p = p[0]
    else:
        p = building.plant_demand_equip['name'][d_pump]
    if building.pump[p]['type']=='VariableSpeed':
            if loop_flow_request>0 and loop_flow_request<building.pump[p]['design_min_flow']:
                demand_bypass_flow = building.pump[p]['design_min_flow'] - loop_flow_request 
                pump_flow = building.pump[p]['design_min_flow']
            elif loop_flow_request>building.pump[p]['design_flow']:
                pump_flow = building.pump[p]['design_flow'] 
            else:
                pump_flow = loop_flow_request 
            PLR = pump_flow/building.pump[p]['design_flow']
            frac_power = building.pump[p]['c1'] + building.pump[p]['c2']*PLR + building.pump[p]['c3']*PLR**2 + building.pump[p]['c4']*PLR**3 
    elif building.pump[p]['type']=='ConstantSpeed':
        try:
            sched = profile[building.pump[p]['schedule']] 
        except KeyError:
            if loop_flow_request>0:
                sched = 1 
            else:
                sched = 0 
        frac_power = sched 
        pump_flow = building.pump[p]['design_flow']*sched 
        demand_bypass_flow = pump_flow - loop_flow_request 
    pump_power = building.pump[p]['design_power']*frac_power #Watts
    shaft_power = building.pump[p]['motor_eff']*pump_power #Watts
    pump_heat = shaft_power + (pump_power - shaft_power)*building.pump[p]['fluid_ineff_frac'] 
    ##put correct flows on demand side branches
    node['demand_flow'][building.plant_demand_nodes['inlet_node'][l]] = pump_flow*building.plant_loop['fluid_density'][l] 
    for j in on_loop:
        in_n = d_equip['inlet'][j]
        out_n = d_equip['outlet'][j] 
        if d_equip['type'][j]=='mixer':
            try:
                in_n[0]
            except TypeError:
                in_n = [in_n]
            node['demand_flow'][out_n] = sum([node['demand_flow'][k] for k in in_n]) 
        elif d_equip['type'][j]=='splitter':
            try:
                out_n[0]
            except TypeError:
                out_n = [out_n]
            for j in on_loop:
                if any([d_equip['inlet'][j] == k for k in out_n]) and d_equip['bypass'][j]:
                    node['demand_flow'][d_equip['inlet'][j]] = demand_bypass_flow*building.plant_loop['fluid_density'][l] # convert back to mass flow
                    break
        else:
            node['demand_flow'][out_n] = node['demand_flow'][in_n] 
    node['supply_flow'][building.plant_supply_nodes['inlet_node'][l]] = node['demand_flow'][building.plant_demand_nodes['outlet_node'][l]] # connect demand half loop to supply half loop
    return node,pump_power,pump_heat


'''
This deviates from EnergyPlus
This collects all building heating and cooling water loops 
and converts to a load on the district heating/cooling water loops
using the district heating/cooling supply temperature
'''

from eagers.building.process.flow_resolver_demand import flow_resolver_demand
from eagers.building.process.demand_loop import demand_loop

def net_water_loop_loads(building,profile,plant_nodes,T_mains,T_supply,dt):   
    Cp_water = 4186  #J/kg*K
    nl = len(T_supply)
    Q_hot = [0 for i in range(nl)]
    Q_cold = [0 for i in range(nl)]
    net_pumps = 0
    for i  in range(len(building.plant_loop['name'])):
        l = building.plant_loop['loop_order'].index(i)
        n = building.plant_demand_nodes['outlet_node'][l] 
        plant_nodes,pump_power,pump_heat = flow_resolver_demand(building,plant_nodes,profile,l,T_mains) 
        plant_nodes = demand_loop(building,profile,T_mains,plant_nodes,l) 
        loop_flow_rate = plant_nodes['demand_flow'][n]
        if loop_flow_rate>0:
            T_return = plant_nodes['demand_temperature'][n] + pump_heat/(Cp_water*loop_flow_rate)
            net_pumps +=pump_power
            T_treat = building.plant_loop['exit_temperature'][l]
            if building.plant_loop['type'][l].lower() == 'heating':
                tank_heat = (T_treat - T_supply[l])*Cp_water*building.plant_loop['loop_volume'][l]/2*building.plant_loop['fluid_density'][l] #energy in J
                Q_hot[l] =  max(0,loop_flow_rate*Cp_water*(T_treat-T_return) + tank_heat/dt) #power in W 
            elif building.plant_loop['type'][l].lower() == 'cooling':
                tank_cool = (T_supply[l] - T_treat)*Cp_water*building.plant_loop['loop_volume'][l]/2*building.plant_loop['fluid_density'][l] #energy in J
                Q_cold[l] = max(0,loop_flow_rate*Cp_water*(T_return-T_treat) + tank_cool/dt) #power in W
    return Q_hot, Q_cold, net_pumps


from eagers.solver.ecos_qp import ecos_qp

def dispatch_network(names,subnet,production,demand,request,marginal,capacity,constrained):
    '''Performs a quadratic optimization to find the shortage/excess of demand on the network
    and the most cost effective node to add energy production
    If constrained is true it treats any buildings as constant loads, otherwise it considers building flexibility'''
    qp = build_net_optimization(names,subnet,production,demand,request,marginal,capacity,constrained)
    x, feasible = ecos_qp(qp['h'], qp['f'], qp['a'], qp['b'], qp['a_eq'], qp['b_eq'], qp['ub'], qp['lb'], qp['x_keep'], qp['r_keep'], qp['req_keep'])
    if feasible == 0:
        if constrained:
            ##TODO return imbalance by node
            thermal = sum(x[-len(subnet['nodes']):]) #sum of slack nodes is the net energy imbalance on the network
        else:
            building_thermal = x[:len(names['buildings'])]
            thermal = [0 for i in range(len(subnet['nodes']))]
            for i in range(len(names['buildings'])):
                for n in range(len(subnet['nodes'])):
                    if 'buildings' in subnet and len(subnet['buildings'][n])>0:
                        if i in subnet['buildings'][n]:
                            thermal[n] += building_thermal[i]*1000
    else:
        if constrained:
            thermal = 0 #no energy imbalance to redistribute
        else:
            thermal = [j*1000 for j in request['nominal']] #stick with original guess from optimization (will result in imbalance somewhere)
        print('Simulation error: Cannot find feasible network solution.')
    return thermal 

def build_net_optimization(names,subnet,production,demand,request,marginal,capacity,constrained):
    x_l = len(names['buildings']) + sum([2 if s == 'dual' else 1 for s in subnet['line']['dir']]) + len(subnet['nodes']) #building thermal, line flow , and slack states
    nn = len(subnet['nodes'])
    qp = {}
    qp['h'] = [0 for j in range(x_l)]
    qp['f'] = [0 for j in range(x_l)]
    qp['x_keep'] = [True for j in range(x_l)]
    qp['a'] = []
    qp['b'] = []
    qp['r_keep'] = []
    qp['a_eq'] = [[0 for j in range(x_l)] for i in range(nn)]
    qp['b_eq'] = [0 for i in range(nn)]
    qp['req_keep'] = [True for j in range(nn)]
    qp['lb'] = [0 for j in range(x_l)]
    qp['ub'] = [0 for j in range(x_l)]
    for n in range(nn):
        if 'buildings' in subnet and len(subnet['buildings'][n])>0:
            ##TODO consider normalizing by building water loop thermal capacity rather than nominal load
            for j in subnet['buildings'][n]:
                qp['a_eq'][n][j] = request['nominal'][j]
                if constrained:
                    qp['lb'][j] = 1
                    qp['ub'][j] = 1 + 1e-8
                else:
                    qp['h'][j] = 1
                    qp['lb'][j] = request['minimum'][j]/request['nominal'][j]
                    qp['ub'][j] = request['maximum'][j]/request['nominal'][j]
    ind = len(names['buildings'])
    node_names = [subnet['nodes'][j][0] for j in range(nn)]
    for i in range(len(subnet['line']['dir'])):
        j = node_names.index(subnet['line']['node1'][i])
        k = node_names.index(subnet['line']['node2'][i])
        qp['a_eq'][j][ind] = -1 #energy leaving node 1
        qp['a_eq'][k][ind] = subnet['line']['eff'][0] #energy arriving at node 2
        qp['ub'][ind] = subnet['line']['limit'][0]
        ind += 1
        if subnet['line']['dir'] == 'dual':
            qp['a_eq'][k][ind] = -1 #energy leaving node 2
            qp['a_eq'][j][ind] = subnet['line']['eff'][1] #energy arriving at node 1
            qp['ub'][ind] = subnet['line']['limit'][1]
            ind += 1
    for n in range(nn):
        qp['a_eq'][n][ind] = 1 #slack variable for production at this node
        qp['b_eq'][n] = production[n]
        if subnet['nodes'][n][0] in list(demand.keys()):
            qp['b_eq'][n] -= demand[subnet['nodes'][n][0]]
        qp['f'][ind] = marginal[n]
        qp['ub'][ind] = capacity[n]
        if not constrained:
            qp['f'][ind] += 1e3 #add cost so it uses building flexibility first
        ind += 1
    return qp


from eagers.building.basic.psychometric import psychometric

import numpy as np

def reset_loop_param(building,T,w,air_nodes,plant_nodes,weather,date):
    sn = len(building.plant_supply_nodes['name'])
    dn = len(building.plant_demand_nodes['name'])
    plant_nodes['load'] =  [0 for i in range(dn)] #reset to zero, and put in loads from demand side as you get through equipment
    plant_nodes['demand_flow'] = [0 for i in range(dn)] #reset to zero, and put in loads from demand side as you get through equipment
    plant_nodes['supply_flow'] = [0 for i in range(sn)]
    sn = len(building.air_supply_nodes['name'])
    dn = len(building.air_demand_nodes['name'])
    air_nodes['supply_T_set'] = [0 for i in range(sn)]
    air_nodes['supply_flow'] = [0 for i in range(sn)]
    air_nodes['demand_T_set'] = [0 for i in range(dn)]
    air_nodes['demand_flow'] = [0 for i in range(dn)]
    
    ###pre-compute some ambient conditions
    sig = 5.67e-8 # W/m^2*K^4 Stephan-Boltzman const
    sky_emissivity = (0.787 + 0.764*np.log((weather['t_dewp']+273)/273))*(1 + 0.0224*weather['opq_cld'] - 0.0035*weather['opq_cld']**2 + 0.00028*weather['opq_cld']**3)
    ir_intensity = sky_emissivity*sig*(weather['t_dryb']+273)**4 #W/m^2
    T['sky'] = (ir_intensity/sig)**.25 - 273
    T['ground'] = building.ground_temperatures[date.month-1]
    w['air'] = psychometric({'P':weather['pres_pa'],'Tdp':weather['t_dewp']},'w')
    T['air_zone'] = [weather['t_dryb'] - 0.0065*h for h in building.ctf['z_height']] #outdoor air temp next to each zone
    T['air_surface'] =[weather['t_dryb'] - 0.0065*h for h in building.ctf['s_height']] #outdoor air temp next to each exterior surface
    T['exterior'] = [j for j in T['air_surface']]
    for i in range(len(building.surfaces['exterior']['name'])):
        if building.surfaces['exterior']['boundary'][i]=='ground':
            T['exterior'][i] = T['ground']
    return T,w,air_nodes,plant_nodes


def area_weighted_temperature(building,Tzone):
    nz = len(building.zones['name'])
    hvac = [0 for i in range(nz)]
    for i in building.zone_controls:
        if building.zone_controls[i]['type']=='Thermostat':
            z = building.zones['name'].index(building.zone_controls[i]['zone'])
            hvac[z] = 1
    A = [building.zones['floor_area'][z]*building.zones['multiplier'][z]*hvac[z] for z in range(nz)]
    T = sum([Tzone[z]*A[z] for z in range(nz)])/sum(A)
    return T


"""Logic for updating energy purchase costs.

Functions:
update_cost
interp_scale_cost
"""

import datetime as dt

from eagers.config.units import kWh_per_MMBtu_natgas
from eagers.basic.get_data import interp_data


def update_cost(date, gen, market):
    """Update energy purchase costs for the current time.

    Positional arguments:
    date - (list of datetime) Dates for horizon (current time not
        included).
    gen - (list of dict) Generator information.
    market - Market information.
    """
    n_s = len(date)
    n_g = len(gen)
    source = [None] * n_g
    scale_cost = [[] for i in range(n_g)]
    market_rate = None
    # Get the rate for each utility.
    # Assumes there is only one utility for each network type.
    for i, g in enumerate(gen):
        if g['type'] == 'Utility':
            source[i] = g['source']
            utility = g
            if 'sum_start_month' in utility:
                # Ensure summer and winter seasons don't start on the same day.
                yr = date[0].year
                sum_start = dt.datetime(yr, utility['sum_start_month'], utility['sum_start_day'])
                if utility['win_start_month'] > utility['sum_start_month'] or \
                    (utility['win_start_month'] == utility['sum_start_month'] \
                        and utility['win_start_day'] > utility['sum_start_day']):
                    yr +=1
                win_start = dt.datetime(yr, utility['win_start_month'], utility['win_start_day'])
                # Whether the winter pricing season starts this year, as
                rate = []
                for d in date:
                    h = d.hour
                    day = d.weekday()
                    if d > sum_start and d <= win_start:
                        j = utility['sum_rate_table'][day][h] - 1
                        rate.append(utility['sum_rates'][j][0])
                    else:
                        j = utility['win_rate_table'][day][h] - 1
                        rate.append(utility['win_rates'][j][0])
                scale_cost[i] = rate  # Utility rate in $/kWh.
            else:
                # Time series utility.
                ut_yr = utility['timestamp'][0].year
                d_yr = date[0].year
                diff = dt.datetime(ut_yr,1,1) - dt.datetime(d_yr,1,1)
                date_num=[(d+diff).replace(tzinfo=dt.timezone.utc).timestamp() for d in date]
                j = 0
                n_u = len(utility['timestamp'])
                while j<n_u-2 and utility['timestamp'][j+1]<date[0]+diff:
                    j+=1
                jj = j+1
                while jj<n_u-1 and utility['timestamp'][jj]<date[-1]+diff:
                    jj+=1
                
                i_list = [i for i in range(j,jj)]
                if jj==len(utility['timestamp'])-1:
                    i_list.append(0)
                    diff = dt.datetime(ut_yr-1,1,1) - dt.datetime(d_yr,1,1)
                    while i_list[-1]<n_u-1 and utility['timestamp'][i_list[-1]]<date[-1]+diff:
                        i_list.append(i_list[-1]+1)
                else:
                    i_list.append(jj)
                conversion = 1
                if source[i] in ('ng', 'diesel'):
                    # Convert gas rate from $/MMBTU to $/kWh.
                    conversion = 1 / kWh_per_MMBtu_natgas
                # Interpolate rate.
                scale_cost[i] = interp_scale_cost(utility, conversion, i_list, date_num)
        elif g['type']=='Tradepoint': 
            #TODO update tradpoint coefficients b0,b1 and s0,s1
            six_param = g['six_param']
            scale_cost[i] = [[six_param] for i in range(n_s)]
        elif g['type']=='Market': #Market varaible/bids is empty during initialization
            if market_rate == None:
                market_rate = [[]]
                m_num = 0
            else:
                market_rate.append([])
                m_num += 1
            na = len(market['award']['time'])
            awarded = 0
            for j in range(na):
                if market['award']['time'] > date[0]:
                    # Identify how many awards have already passed.
                    awarded += 1
            n = na - awarded
            for d in date:
                while n < na and d < market['award']['time'][n]:
                    n += 1
                market_rate[m_num].append(market['award']['price'][m_num][n])
            scale_cost[i] = market_rate[m_num]

    # Match each utility-fueled generator with its corresponding utility
    # cost, if a corresponding utility exists.
    for i, g in enumerate(gen):
        util_fueled = ('ElectricGenerator', 'CombinedHeatPower', 'Chiller',
            'Heater', 'CoolingTower', 'Electrolyzer', 'HydrogenGenerator')
        if g['type'] in util_fueled:
            try:
                util_index = source.index(g['source'])
            except ValueError:
                # No utility (don't scale costs).
                util_index = None
                scale_cost[i] = [1] * n_s
            if util_index is not None:
                scale_cost[i] = scale_cost[util_index]
    return scale_cost


def interp_scale_cost(utility, conversion, i_list, date_num):
    x_vals = []
    for i in i_list:
        t_add = utility['timestamp'][i]
        if i <i_list[0]:
            # Add year.
            t_add += dt.timedelta(days = 365)
        x_vals.append((t_add).replace(tzinfo=dt.timezone.utc).timestamp())
    y_vals = {}
    if 'SP' in utility:
        # Spot market.  Get spot market price signal.
        y_vals['rate']=[utility['SP'][t] * conversion for t in i_list]
    else:
        y_vals['rate']=[utility['rate'][t] * conversion for t in i_list]
    dtq1 = x_vals[1]-x_vals[0]
    scale = interp_data(x_vals, y_vals, date_num, ['rate'], None, dtq1)['rate']
    return scale



"""Logic for checking whether AC/DC conversion capabilities are properly
outfitted for a project.

Functions:
check_ac_dc
"""

def check_ac_dc(components, buildings, test_data):
    """Check whether AC/DC conversion capabilities are correctly
    configured for the given set of components, buildings, and test
    data.
    """
    has_dc = False
    has_ac = False
    has_ac_dc = False
    need_ac = False
    need_dc = False

    for comp in components:
        if comp['type'] == 'ACDCConverter':
            has_ac_dc = True
        else:
            if 'output' in comp:
                if 'direct current' in comp['output']:
                    has_dc = True
                if 'electricity' in comp['output']:
                    has_ac = True
            if 'source' in comp:
                if 'direct_current' in comp['source']:
                    has_dc = True
                if 'electricity' in comp['source']:
                    has_ac = True
        if comp['type'] == 'HydroStorage':
            has_ac = True
    if ('nodedata' in test_data \
            and len(test_data['nodedata'].tablenames) > 0
            and any('e' in test_data['nodedata'].colnames(tn)
                for tn in test_data['nodedata'].tablenames)):
        need_ac = True
    if ('nodedata' in test_data \
            and len(test_data['nodedata'].tablenames) > 0
            and any('dc' in test_data['nodedata'].colnames(tn)
                for tn in test_data['nodedata'].tablenames)):
        need_dc = True
    if buildings:
        need_ac = True
        for b in buildings:
            if 'dc_loads' in b:
                need_dc = True

    if not has_ac_dc and (need_ac and not has_ac) or (need_dc and not has_dc):
        # Add default AC/DC converter model to list of components.
        qpform = dict(
            # First state is AC power transferred to DC power.
            # Second state is DC power transferred to AC power.
            states=[['a', 'b']],
            output=dict(e=[[-1, 1]],dc=[[1-1e-4, -1]]),#Prevent infinite backward/forward transfer in ideal
            a=dict(ub=1e12, lb=0, f=0, h=0,),
            b=dict(ub=1e12, lb=0,  f=0,  h=0,),        )
        components.append(qpform)    
    return components



'''
overides temperature setpoints at specific nodes, e.g. cold air supply
'''
from eagers.building.components.fan_calc import fan_calc
from eagers.building.basic.psychometric import psychometric

def manager_override(building,profile,node,setpoint,air,fresh_air,fan_avail):
    k = [i for i,x in enumerate(building.manager['node']) if x ==node]
    if len(k)>0:
        for i in k:
            if building.manager['type'][i]=='OutdoorAirReset':
                if building.manager['low_temp'][i]<building.manager['high_temp'][i]:
                    if fresh_air['Tdb'] < building.manager['low_temp'][i]:
                        setpoint = building.manager['setpoint_at_low_temp'][i]
                    elif fresh_air['Tdb']>building.manager['high_temp'][i]:
                        setpoint = building.manager['setpoint_at_high_temp'][i]
                    else:
                        setpoint = building.manager['setpoint_at_low_temp'][i] - (fresh_air['Tdb'] - building.manager['low_temp'][i])/(building.manager['high_temp'][i] - building.manager['low_temp'][i])*(building.manager['setpoint_at_low_temp'][i] - building.manager['setpoint_at_high_temp'][i])
                else:
                    setpoint = 0.5*(building.manager['setpoint_at_low_temp'][i] + building.manager['setpoint_at_high_temp'][i])
            elif building.manager['type'][i]=='Scheduled':
                setpoint = profile[building.manager['schedule'][i]]
            elif building.manager['type'][i]=='MixedAir':
                if air['m_dot']>0:
                    for x in building.fans:
                        if building.fans[x]['inlet']==building.manager['fan_inlet_node'][i]:
                            fan_name = x
                            break
                    _,_,Q_to_air = fan_calc(building.fans[fan_name],air,fan_avail[fan_name])
                    #cooling in zone calculation is m*Cp*dT, not h2-h1, so need to ensure sensible load = m*Cp*(Tsupply -Tzone)
                    spec_heat_supply = 1006 + 1860*air['w'] #J/kg*K
                    setpoint = setpoint - Q_to_air/(air['m_dot']*spec_heat_supply)
            elif building.manager['type'][i]=='FollowOutdoorAirTemperature':
                if building.manager['ref_temp_type'][i]=='OutdoorAirWetBulb' or building.manager['ref_temp_type'][i]=='OutdoorWetBulb':
                    setpoint = psychometric(fresh_air,'Twb') + building.manager['offset_temperature'][i]
                elif building.manager['ref_temp_type'][i]=='OutdoorAirDryBulb' or building.manager['ref_temp_type'][i]=='OutdoorDryBulb':
                    setpoint = fresh_air['Tdb'] + building.manager['offset_temperature'][i]
                setpoint = max(building.manager['min_temperature'][i],min(building.manager['max_temperature'][i],setpoint))
            elif building.manager['type'][i]=='SingleZone:Reheat':
                #handled elsewhere
                pass
            elif building.manager['type'][i]=='SingleZone:Humidity:Minimum':
                #need to add for hospital see page 1396
                pass
            elif building.manager['type'][i]=='SingleZone:Humidity:Maximum':
                #need to add for hospital
                pass
            else:
                print('need more HVAC manager types')
    return setpoint


"""Simulation run logic.

Functions:
run_eagers_core - read from excel files, pre-load dictionaries and call run_project
run_project - Initializes data structures and calls run_simulation().
run_simulation - Low-level simulation run logic.
load_actual_data
simulate_response
"""

import datetime

from eagers.read.excel_interface import ProjectTemplateReader, TestDataTemplateReader
from eagers.setup.preload import preload
from eagers.read.read_test_data import read_test_data
from eagers.read.read_network import read_network
from eagers.read.read_market import read_market
from eagers.building.process.load_actual_building_data import load_actual_building_data
from eagers.building.process.forecast_dr_capacity import forecast_dr_capacity
from eagers.config.simulation import DEFAULT_INITIAL_SOC
from eagers.basic.build_time_vector import build_time_vector
from eagers.basic.get_data import get_data
from eagers.basic.result_template import result_template
from eagers.basic.auto_set_startdate import auto_set_startdate
from eagers.simulate.plant_response import plant_response
from eagers.forecasting.calculate_fits import calculate_fits
from eagers.forecasting.update_forecast import update_forecast
from eagers.forecasting.hydro_forecast import hydro_forecast
from eagers.forecasting.water_year_forecast import water_year_forecast
from eagers.setup.automatic_ic import automatic_ic
from eagers.setup.initialize_observer import initialize_observer, initialize_building_observer
from eagers.solver.dispatch_loop import dispatch_loop
from eagers.solver.generate_bids import generate_bids
from eagers.update.dispatch_record import dispatch_record, predict_record
from eagers.simulate.renewable_output import renewable_output
from eagers.update.update_cost import update_cost
from eagers.update.update_market import update_market
from eagers.update.update_observer import update_observer
from eagers.write.result_file import new_result_file, result_file_dimensions, append_step_solution
from eagers.plot.dispatch_step import new_figure, plot_component_dispatch


def demo():
    run_eagers_core('default_project', 'default_testdata', plot_step=True) #Run default project

def demo_result():
    from eagers.plot.dispatch_result_excel import plot_dispatch_result
    plot_dispatch_result('default_project', 'default_testdata')

def demo_building():
    from eagers.building.validate_building import validate_building
    validate_building(build = 1,climate = '4A', start_day = 1, duration = 10, horizon = 0, mode = 'build',warmup = 'new')

def run_eagers_core(project_filename, testdata_filename,  *, plot_step=True):
    """Load from excel then run a project 

        Keyword arguments:
        plot_step - (bool) (Default: True) Whether dispatch should be
            plotted at each simulation iteration.
    """
    
    proj = ProjectTemplateReader.read_userfile(project_filename)
    proj['test_data'] = read_test_data(TestDataTemplateReader.read_userfile(testdata_filename))
    read_network(proj['plant']['network']) 
    proj['plant']['market'] = read_market(None)
    if proj['options']['start_date'] is None:
        if len(proj['plant']['building'])>0:
            proj['options']['start_date'] = proj['plant']['building'][0].sim_date[0]
        else:
            proj['options']['start_date'] = auto_set_startdate(proj['test_data'])
    proj['preload'] = preload(proj['plant'], proj['test_data'] ,proj['options'])
    run_project(proj, plot_step)

def run_project(proj, plot_step):
    """Run the given project. Initializes data structures and then
    passes them on to run_simulation().

    Positional arguments:
    proj - (Project) Project to run.
    plot_step - (bool) Whether dispatch should be plotted at each
        simulation iteration.
    """
    date = [proj['options']['start_date']]
    date_v = build_time_vector(proj['options']['start_date'], proj['options'], to_timedelta=True)
    
    building_observer,zones,pl = initialize_building_observer(proj['plant']['building'],proj['test_data']['weather'],date)    
    names, dimensions = result_file_dimensions(proj['preload'],proj['plant'],date_v,zones,pl)
    new_result_file(proj['name'], names, dimensions, zones, pl)

    # Create typical day fits if necessary.
    proj['hist_prof'] = calculate_fits(
        proj['test_data'], proj['options'], proj['preload']['subnet'])
    market = proj['plant']['market']
    fluid_loop = create_fluid_loop(proj['plant'])
    
    scale_cost = update_cost(date, proj['preload']['gen_qp_form'], market)
    ic, data_t0, _ = automatic_ic(
        proj['preload']['gen_qp_form'], proj['plant']['building'], fluid_loop,
        market, proj['preload']['subnet'], date, proj['preload']['one_step'], proj['options'],
        proj['test_data'], building_observer, scale_cost)
    
    all_data_nodes = count_nodes(proj['preload']['subnet'],proj['test_data'])
    proj['observer'] = initialize_observer(proj['preload']['gen_qp_form'], proj['preload']['subnet'], fluid_loop,all_data_nodes, names, zones, pl, proj['test_data'], 
                            building_observer, ic, DEFAULT_INITIAL_SOC, date, data_t0, proj['name'])
    
    # If October 1st, run a yearly forecast for hydrology.
    proj['observer'] = water_year_forecast(proj['preload']['gen_qp_form'], proj['plant']['building'],
         fluid_loop, proj['observer'], all_data_nodes, names, zones, pl,
        market, proj['preload']['subnet'], proj['options'], date_v, proj['test_data'])
    run_simulation(data_t0, date,  proj['test_data'], proj['preload'],
        proj['plant']['building'], fluid_loop, all_data_nodes, names, zones, pl,
        proj['observer'], market,proj['options'], proj['name'], plot_step)


def run_simulation(
        data_t0, date, test_data, preload, building,
        fluid_loop, all_data_nodes, names, zones, pl,
        observer, market, options, project_name, plot_step):
    """Low-level simulation run logic.

    Positional arguments:
    ...
    plot_step - (bool) Whether to plot at each simulation step.
    """
    # Break up preload into smaller variables.
    gen = preload['gen_qp_form']
    subnet = preload['subnet']
    
    # Get figure and axes for plotting simulation step results.
    if plot_step:
        fig, axs = new_figure(project_name, gen, subnet)

    # Set up vector of time interval.
    date = build_time_vector(options['start_date'], options, to_timedelta=True)

    # Simulation loop.
    if options['method'] == 'planning':
        num_steps = int(options['interval'] * 24 / (options['resolution']*options['horizon']) + 1)
    else:
        num_steps = int(options['interval'] * 24 / options['resolution'] + 1)
    timer =  []
    prediction = None
    predicted = result_template(all_data_nodes,names, zones, pl)
    for k in predicted['building']:
        del predicted['building'][k]['supply']
        del predicted['building'][k]['return_']
    dispatch = result_template(all_data_nodes,names, zones, pl)
    while date[0] < options['start_date']+ datetime.timedelta(days=options['interval']):
        market = update_market(gen, market, date)# Assign any previous market bids if a bidding period has closed.
        
        observer = water_year_forecast(
            gen, building, fluid_loop, observer, all_data_nodes, names, zones, pl, market, subnet, options, date,test_data)# If October 1st, run a yearly forecast for hydrology.
        forecast = update_forecast(options, date[1:], test_data, subnet)
        if 'weather' in forecast and 'dir_norm_irr' in forecast['weather']:
            forecast['renewable'] = [renewable_output(gen[i], date[1:], forecast['weather']['dir_norm_irr']) \
                                    if gen[i]['type'] in ['Renewable','Solar','Wind'] and gen[i]['enabled'] \
                                    else [] for i in range(len(gen))]
        forecast['building'], _, _, _ = forecast_dr_capacity(building, observer, forecast['weather'], date)
        forecast = hydro_forecast(forecast, test_data, date[1:], subnet, observer, options)

        solution, _ = dispatch_loop(gen, observer, market, subnet, names, preload['op_mat_a'], 
                                    preload['op_mat_b'], preload['one_step'], options, date, forecast, prediction)
        timer.append(solution['timer'])
        
        predict_record(gen, predicted, observer, market, subnet, date, forecast, solution, project_name)
        market = generate_bids(gen, solution['dispatch'], date, market)  # Generate new bids for next time steps.

        if options['method'] == 'planning':
            actual_data = forecast# Assumes forecast is perfect (no disparity).
        elif  options['method'] in ('dispatch', 'control'):   
            actual_data = load_actual_data(options, observer, test_data, 
                forecast, date, building, subnet)         
        plant_response(gen, building, fluid_loop, observer, market, subnet, names, actual_data,
                       preload, options, forecast, date, solution)

        dispatch_record(gen, dispatch, observer, market, subnet, actual_data, solution, project_name)
        date = update_observer(gen, subnet, observer, date, options, solution, actual_data, forecast)
        prediction = next_prediction(gen,observer,solution['dispatch'])
        del solution['dispatch']
        append_step_solution(project_name, solution)# Write result of simulation step to HDF5 file.)
        # Plot.
        if plot_step:
            try:
                plot_component_dispatch(
                    axs, observer['history'], observer['future'], gen, subnet)
            except:
                # Stop plotting at every step so the simulation can
                # continue.
                plot_step = False


def load_actual_data(options, observer, test_data, forecast, date, buildings, subnet):
    # Count forward 1 step, rounded to nearest second.
    d_now = [date[0] + datetime.timedelta(hours=options['resolution'])]
    actual_data = get_data(test_data, d_now, subnet['network_names'])
    actual_data['building'] = load_actual_building_data(buildings,observer,actual_data['weather'],forecast['building'],d_now,options['resolution']*3600)
    return actual_data


def next_prediction(gen,observer,forecast):
    n_g = len(gen)
    prediction = [[] for j in range(n_g)]
    for i in range(n_g):
        if 'stor' in gen[i]:
            prediction[i] = [observer['stor_state'][i]]
        else:
            prediction[i] = [observer['gen_state'][i]]
        prediction[i].extend(forecast[i][2:])
        prediction[i].append(forecast[i][-1])
    return prediction

def create_fluid_loop(plant):
    fluid_loop = {}
    if len(plant['fluid_loop'])>0:
        kys = list(plant['fluid_loop'][0].__dict__.keys())
        for k in kys:
            fluid_loop[k] = []
        for j in range(len(plant['fluid_loop'])):
            for k in kys:
                fluid_loop[k].append(getattr(plant['fluid_loop'][j],k))
    else:
        fluid_loop['name'] = []
    return fluid_loop

def count_nodes(subnet,test_data):
    all_data_nodes = []
    for net in subnet['network_names']:
        if net in test_data['nodedata_network_info']:
            all_nodes = []
            net_nodes = []
            data_nodes = [test_data['nodedata_network_info'][net][i]['node'] for i in range(len(test_data['nodedata_network_info'][net]))]
            for i in range(len(subnet[net]['nodes'])):
                ag_nodes = subnet[net]['nodes'][i]
                all_nodes.extend(ag_nodes)
                net_nodes.extend([i for j in range(len(ag_nodes))])
            for j in range(len(data_nodes)):
                n = net_nodes[all_nodes.index(data_nodes[j])]
                if 'demand' in test_data['nodedata_network_info'][net][j]:
                    subnet[net]['load'][n].append(test_data['nodedata_network_info'][net][j]['demand'])
            all_data_nodes.extend(data_nodes)
    return all_data_nodes


'''
mode 1 (empty plant_nodes, only heating coils): Given inlet temperatures and zone load request, for given air_flow find water flow that meets load. If water flow saturates, increase air flow until load is met or air flow saturates
mode 2: Given water and air inlet flows and temperature, find air outlet
Detailed water coil applies only to cooling, only mode 2
'''
from eagers.building.basic.psychometric import psychometric
from eagers.building.basic.two_way_interp import two_way_interp
from eagers.building.components.detailed_water_coil import detailed_water_coil

import numpy as np
import copy

def water_coil(building,name,var1,var2,var3):
    if name in building.coils_heating:
        coil  = building.coils_heating[name]
        mode = 'heating'
    else:
        coil = building.coils_cooling[name]
        mode = 'cooling'

    n = building.plant_demand_equip['name'].index(name)   
    out = building.plant_demand_equip['outlet'][n] 
    if type(var3) is dict:
        solve4 = 'water'
        air_in = {}
        for f in list(var1.keys()):
            air_in[f] = copy.copy(var1[f])
        T_supply = copy.copy(var2)
        in_n = building.plant_demand_equip['inlet'][n]
        Tw_in = var3['demand_temperature'][in_n]
        Cp_air = 1006 + 1860*air_in['w'] #J/kg*K
        if mode=='heating':
            sensible_load  = air_in['m_dot']*Cp_air*(T_supply - air_in['Tdb'])
        elif mode=='cooling':
            sensible_load  = air_in['m_dot']*Cp_air*(air_in['Tdb'] - T_supply)
    else:
        solve4 = 'air'
        air_in = {}
        air_in['Tdb']  = var3[0]
        air_in['w']  = var3[1]
        m_a_max = var3[2]
        Cp_air = 1006 + 1860*air_in['w'] #J/kg*K
        air_in['m_dot'] = var2/(Cp_air*(coil['air_outlet_temperature']-var1))
        sensible_load = air_in['m_dot']*Cp_air*(coil['air_outlet_temperature'] - air_in['Tdb'])
        air_in['m_dot'] = min(air_in['m_dot'],m_a_max)
        air_in['h'] = psychometric(air_in,'h')
        Tw_in = coil['water_inlet_temperature']
        T_supply = var3[0]
    if coil['type']=='Cooling:Water:DetailedGeometry' or coil['type']=='Heating:Water:DetailedGeometry':
        air_out,m_w,actual_load = detailed_water_coil(coil,air_in,sensible_load,Tw_in,T_supply)
    elif coil['type']=='Heating:Water':
            air_out,m_w,actual_load = heating_water_coil(coil,air_in,sensible_load,Tw_in,T_supply,solve4)
    elif coil['type']=='Cooling:Water':
            air_out,m_w,actual_load = cooling_water_coil(coil,air_in,sensible_load,Tw_in,T_supply)
    return air_out,m_w,actual_load,out

def heating_water_coil(coil,air_in,sensible_load,Tw_in,T_supply,solve4):
    #sensible energy transfer only
    air_out = {}
    for f in list(air_in.keys()):
        air_out[f] = copy.copy(air_in[f])
    air_density = 1.204 #kg/m^3
    if air_in['m_dot']>1e-3*coil['rated_air_flow']*air_density and sensible_load/air_in['m_dot']>1 and coil['rated_water_flow']>0:
        den_water = 995 #kg/m^3
        Cp_air = 1006 + 1860*air_in['w'] #J/kg*K
        Cp_water = 4186 #J/kg*K
        Ta_in = air_in['Tdb']
        Tw_in0 = coil['water_inlet_temperature']
        Ta_in0 = coil['air_inlet_temperature']
        Tw_out = min(coil['water_outlet_temperature'],Tw_in - 1)
        m_a_rated = coil['rated_air_flow']*air_density
        m_a = air_in['m_dot']
        m_w_rated = coil['rated_water_flow']*den_water #convert m^3/s to kg/s
        hA_w0,hA_a0,n_f = w_coil_param(coil)
        x_a = 1 + 4.769e-3*(Ta_in - Ta_in0)
        x_w = 1 + (.014/(1 + .014*Tw_in0))*(Tw_in - Tw_in0)
        if solve4=='air':
            ##line search to solve for water flow that gives correct sensible load (page 789 & 960 of the manual)
            ##water flow rate and water outlet temperature change
            n = 10
            h_out = [0 for i in range(n)]
            m_w_min = max(1e-8*m_w_rated,min(m_w_rated,.02*sensible_load/(Cp_water*(Tw_in - Tw_out))))
            flow = np.logspace(np.log10(m_w_min),np.log10(m_w_rated),n)
            flow = [float(x) for x in flow]
            for i in range(n):
                T_out = dry_coil(x_w,flow[i],m_w_rated,hA_w0,x_a,m_a,m_a_rated,hA_a0,n_f,Cp_air,Ta_in,Tw_in)
                h_out[i] = psychometric({'Tdb':T_out,'w':air_out['w']},'h')
            actual_load = [air_out['m_dot']*(h_out[i] - air_in['h']) for i in range(n)] #heat added in W
            if actual_load[-1]<sensible_load:
                m_w = m_w_rated
            else:
                m_w = two_way_interp(sensible_load,actual_load,flow)
            ## If increasing water flow was insufficient, solve for air flow to meet load
            flow = np.linspace(air_out['m_dot'],m_a_rated,n)
            flow = [float(x) for x in flow]
            for i in range(n):
                T_out = dry_coil(x_w,m_w,m_w_rated,hA_w0,x_a,flow[i],m_a_rated,hA_a0,n_f,Cp_air,Ta_in,Tw_in)
                h_out[i] = psychometric({'Tdb':T_out,'w':air_out['w']},'h')
            actual_load = [flow[i]*(h_out[i] - air_in['h']) for i in range(n)] #heat added in W
            if actual_load[-1]<sensible_load:
                air_out['m_dot'] = m_a_rated
            else:
                air_out['m_dot'] = two_way_interp(sensible_load,actual_load,flow)
        elif solve4=='water': #if m_w == 0 %first guess solving for m_w
            ##line search to solve for water flow that gives correct sensible load (page 789 & 960 of the manual)
            ##water flow rate and water outlet temperature change
            n = 10
            T_out = [0 for i in range(n)]
            m_w_min = max(1e-8*m_w_rated,min(m_w_rated,.002*sensible_load/(Cp_water*(Tw_in - Tw_out))))
            flow = np.logspace(np.log10(m_w_min),np.log10(m_w_rated),n)
            flow = [float(x) for x in flow]
            for i in range(n):
                T_out[i] = dry_coil(x_w,flow[i],m_w_rated,hA_w0,x_a,m_a,m_a_rated,hA_a0,n_f,Cp_air,Ta_in,Tw_in)
            if T_out[-1]<T_supply:
                m_w = m_w_rated
            else:
                m_w = two_way_interp(T_supply,T_out,flow)
            air_out['Tdb'] = dry_coil(x_w,m_w,m_w_rated,hA_w0,x_a,m_a,m_a_rated,hA_a0,n_f,Cp_air,Ta_in,Tw_in)
            air_out['h'] = psychometric(air_out,'h')
        actual_load = -air_out['m_dot']*(air_out['h'] - air_in['h']) #heat added in W
    else:
        air_out['Tdb'] = T_supply
        air_out['h'] = psychometric(air_out,'h')
        actual_load = 0
        m_w = 0
    return air_out,m_w,actual_load

def w_coil_param(coil):
    #part of UA calculations depending only on inputs
    if 'air_water_convection_ratio' in coil and coil['air_water_convection_ratio']!=None:
        r = coil['air_water_convection_ratio']
        hA_w0 = coil['UA']*((r+1)/r)
        hA_a0 = coil['capacity']/abs(coil['air_outlet_temperature'] - coil['air_inlet_temperature'])
        n_f = r*hA_w0/hA_a0
    else: # need to find root where r and n_f satisfy set of equations
        den_water = 995 #kg/m^3
        Cp_water = 4186 #J/kg*K
        hA_a0 = coil['capacity']/abs(coil['air_outlet_temperature'] - coil['air_inlet_temperature'])
        hA_w0 = coil['rated_water_flow']*den_water*Cp_water #coil['capacity']/abs(coil['water_inlet_temperature'] - coil['water_outlet_temperature'])
        n_f = 1/(hA_a0/coil['UA']- hA_a0/hA_w0) #fin efficiency  
        r = n_f*hA_a0/hA_w0 #air water convection ratio
    return hA_w0,hA_a0,n_f

def dry_coil(x_w,m_w,m_w_rated,hA_w0,x_a,m_a,m_a_rated,hA_a0,n_f,Cp_air,Ta_in,Tw_in):
    #find new UA factor
    Cp_water = 4186 #J/kg*K
    hA_a = x_a*((m_a/m_a_rated)**.8)*hA_a0
    hA_w = x_w*((m_w/m_w_rated)**.85)*hA_w0
    UA = 1/(1/hA_w + 1/(n_f*hA_a))

    #capacitance
    C_air = Cp_air*m_a
    C_w = Cp_water*m_w
    C_min = min(C_air,C_w)
    Z = C_min/max(C_air,C_w)

    #Find outlet temperatures
    NTU = UA/C_min
    effect = 1 - float(np.exp((np.exp(-Z*NTU**.78)-1)/(Z*NTU**(-.22)))) #cross flow effectiveness
    # effect = (1 - exp(-NTU.*(1-Z)))./(Z*exp(-NTU.*(1-Z))) #counter flow effectiveness
    Ta_out = Ta_in - effect*C_min*(Ta_in - Tw_in)/C_air
    # Tw_out = Tw_in + effect*C_min*(Ta_in - Tw_in)/C_w
    return Ta_out

def cooling_water_coil(coil,air_in,sensible_load,Tw_in,T_supply):
    #sensible energy transfer only
    air_out = {}
    for f in list(air_in.keys()):
        air_out[f] = copy.copy(air_in[f])
    m_w = 0
    if air_in['m_dot']>0 and coil['rated_water_flow']>0 and sensible_load/air_in['m_dot']>1 and Tw_in<air_in['Tdb']:
        air_density = 1.204 #kg/m^3
        den_water = 995 #kg/m^3
        Cp_water = 4186 #J/kg*K
        Cp_air = 1006 + 1860*air_in['w'] #J/kg*K
        Ta_in = air_in['Tdb']
        Tw_in0 = coil['water_inlet_temperature']
        Ta_in0 = coil['air_inlet_temperature']
        Tw_out = Tw_in + (coil['water_outlet_temperature'] - coil['water_inlet_temperature'])
        m_a_rated = coil['rated_air_flow']*air_density
        m_a = air_in['m_dot']
        m_w_rated = coil['rated_water_flow']*den_water #convert m^3/s to kg/s
        hA_w0,hA_a0,n_f = w_coil_param(coil)
        MaxQ = m_w_rated*Cp_water*(Ta_in-Tw_in)
        if Tw_in > air_in['Tdb']:
            print('cooling coil water hotter than air inlet') #should never be here except maybe during warm-up?
        else:
            if Tw_in>T_supply:
                T_supply = Tw_in+0.1 #should never be here except maybe during warm-up?
        
            ##loop to solve for wet/dry portion of coil (page 789 & 960 of the manual)
            ##water flow rate and water outlet temperature change
            Tdp = psychometric(air_in,'Tdp')

            dry = False
            if Tdp < Tw_in:
                dry = True #dry because water is not cold enough to condense moisture
            elif Tdp < T_supply:
                dry = True #dry because it does need to get to the dew point
            elif MaxQ<(air_in['m_dot']*(air_in['h'] - psychometric({'Tdp':Tdp,'w':air_in['w']},'h'))):
                dry = True #dry because there is not enough water cooling capcity to reach dew point
            if dry:
                #simulate dry coil
                x_a = 1 + 4.769e-3*(Ta_in - Ta_in0)
                x_w = 1 + (.014/(1 + .014*Tw_in0))*(Tw_in - Tw_in0)
                ##line search to solve for water flow that gives correct sensible load (page 789 & 960 of the manual)
                ##water flow rate and water outlet temperature change
                n = 10
                T_out = [0 for i in range(n)]
                T_out[-1] = dry_coil(x_w,m_w_rated,m_w_rated,hA_w0,x_a,m_a,m_a_rated,hA_a0,n_f,Cp_air,Ta_in,Tw_in)
                if T_out[-1]>T_supply:
                    m_w = m_w_rated
                    air_out['Tdb'] = T_out[-1]
                else: 
                    m_w_min = max(1e-8*m_w_rated,min(m_w_rated,.002*sensible_load/(Cp_water*(Tw_out - Tw_in))))
                    flow = np.logspace(np.log10(m_w_min),np.log10(m_w_rated),n)
                    flow = [float(x) for x in flow]
                    for i in range(n-1):
                        T_out[i] = dry_coil(x_w,flow[i],m_w_rated,hA_w0,x_a,m_a,m_a_rated,hA_a0,n_f,Cp_air,Ta_in,Tw_in)
                    m_w = two_way_interp(T_supply,T_out,flow)
                    flow = np.logspace(np.log10(.85*m_w),np.log10(min(m_w_rated,1.25*m_w)),n)
                    T_out = [dry_coil(x_w,flow[i],m_w_rated,hA_w0,x_a,m_a,m_a_rated,hA_a0,n_f,Cp_air,Ta_in,Tw_in) for i in range(n)]
                    if T_out[-1]>T_supply:
                        flow = np.logspace(np.log10(1.25*m_w),np.log10(min(m_w_rated,2*m_w)),n)
                        T_out = [dry_coil(x_w,flow[i],m_w_rated,hA_w0,x_a,m_a,m_a_rated,hA_a0,n_f,Cp_air,Ta_in,Tw_in) for i in range(n)]
                    elif T_out[0]<T_supply:
                        flow = np.logspace(np.log10(.5*m_w),np.log10(min(m_w_rated,.85*m_w)),n)
                        T_out = [dry_coil(x_w,flow[i],m_w_rated,hA_w0,x_a,m_a,m_a_rated,hA_a0,n_f,Cp_air,Ta_in,Tw_in) for i in range(n)]
                    m_w = two_way_interp(T_supply,T_out,flow)
                    air_out['T'] = T_supply+0
                #air_out['Tdb'] = dry_coil(x_w,m_w,m_w_rated,hA_w0,x_a,m_a,m_a_rated,hA_a0,n_f,Cp_air,Ta_in,Tw_in)
                air_out['h'] = psychometric(air_out,'h')
            else:
                #wet or partially wet coil
                #must at leach reach dew point
                #search to find water mass flow and dry length
                air_mid  = {}
                for f in list(air_in.keys()):
                    air_mid [f] = copy.copy(air_in[f])
                air_mid['Tdb'] = Tdp
                air_mid['h'] = psychometric(air_mid,'h')
                q_dry = air_in['m_dot']*(air_in['h'] - air_mid['h'])
                
                air_out['Tdb'] = T_supply #initial guess
                air_out['w'] = psychometric({'Twb':air_out['Tdb']},'w')
                if air_in['w']<=air_out['w']:
                    air_out['w'] = air_in['w']+0
                condensate_H2O = air_in['m_dot']*(air_in['w'] - air_out['w'])
                latent_power = condensate_H2O*2264.705 # kg/s *kJ/kg
                air_out['m_dot'] = air_in['m_dot'] - condensate_H2O
                air_out['h'] = psychometric(air_out,'h')
                
                q_wet = max(0,air_mid['m_dot']*air_mid['h'] - air_out['m_dot']*air_out['h'] + latent_power)

                m_w_min = max(1.01*q_wet/(Cp_water*(Tdp - Tw_in)),(q_dry+q_wet)/(Cp_water*(air_in['Tdb'] - Tw_in)))
                m_w_min = max(1e-8*m_w_rated,m_w_min)
                if m_w_rated<=m_w_min:
                    m_w = m_w_rated #need all the cooling we can get
                elif q_wet == 0:
                    m_w = m_w_min
                else:
                    mw_r = [m_w_min,m_w_rated]
                    #for a given water flow Tw_mid is known, and the length of dry coil
                    #to get to Tdb can be found. Iterate m_w to get total heat transfer correct
                    min_e = 1
                    n = 10
                    while min_e>1e-3 and (mw_r[1]-mw_r[0])>0.01*(m_w_rated - m_w_min):
                        flow = np.logspace(np.log10(mw_r[0]),np.log10(mw_r[1]),n)
                        flow = [float(x) for x in flow]
                        Tw_mid =[min(Tdp-1e-8,Tw_in + q_wet/(Cp_water*flow[i])) for i in range(n)]
                        Tlmtd = [((Tdp-Tw_mid[i]) - (T_supply-Tw_in))/float(np.log(Tdp-Tw_mid[i]) - np.log(T_supply-Tw_in)) for i in range(n)]
                        UA_wet_needed = [q_wet/Tlmtd[i] for i in range(n)]
                        UA_wet,_ = wet_coil(flow,m_w_rated,hA_w0,m_a,m_a_rated,hA_a0,n_f,Cp_air,Ta_in,Tw_in,Tw_mid,Tdp,Ta_in0,Tw_in0)
                        min_needed = float(np.amin(UA_wet_needed))
                        error = [(UA_wet[i]-UA_wet_needed[i])/min_needed for i in range(n)]
                        min_e = abs(error[0])
                        mw_r = [flow[0],flow[1]]
                        ind = 0
                        for i in range(n-2):
                            if abs(error[i+1]) <min_e:
                                if error[i+1]<0:
                                    mw_r = [flow[i+1],flow[i+2]]
                                else:
                                    mw_r = [flow[i],flow[i+1]]
                                ind = i+1
                    if UA_wet[-1]<UA_wet_needed[ind]:
                        m_w = flow[-1]
                    elif UA_wet[0]>UA_wet_needed[ind]:
                        m_w = flow[0]
                    else:
                        flow = [flow[i] for i in range(n) if UA_wet[i]>0]
                        UA_wet = [UA_wet[i] for i in range(n) if UA_wet[i]>0]
                        if UA_wet_needed[ind]<UA_wet[0]:
                            m_w = flow[0]
                        else:
                            m_w = two_way_interp(UA_wet_needed[ind],UA_wet,flow)
    actual_load = -air_out['m_dot']*(air_out['h'] - air_in['h'])#heat added in W
    return air_out,m_w,actual_load

def wet_coil(flow,m_w_rated,hA_w0,m_a,m_a_rated,hA_a0,n_f,Cp_air,Ta_in,Tw_in,Tw_mid,Tdp,Ta_in0,Tw_in0):
    #find fraction of coil thats dry, then the UA of the wet portion
    Cp_water = 4186 #J/kg*K
    C_air = Cp_air*m_a
    n = 20 #len(flow)
    frac = [0 for i in range(len(flow))]
    UA_wet = [0 for i in range(len(flow))]
    x_a = 1 + 4.769e-3*(Ta_in - Ta_in0)
    hA_a = x_a*((m_a/m_a_rated)**.8)*hA_a0
    for i in range(len(flow)):
        m_w = flow[i]
        f = np.logspace(-2,0,n)
        f = [float(x) for x in f]
        C_w = Cp_water*m_w
        C_min = min(C_air,C_w)
        Z = C_min/max(C_air,C_w)
        x_w = 1 + (.014/(1 + .014*Tw_in0))*(Tw_mid[i] - Tw_in0)
        hA_w = x_w*((m_w/m_w_rated)**.85)*hA_w0
        UA = [f[j]/(1/(hA_w) + 1/(n_f*hA_a)) for j in range(n)]
        NTU = [UA[j]/C_min for j in range(n)]
        effect = [1 - float(np.exp((np.exp(-Z*NTU[j]**.78)-1)/(Z*NTU[j]**(-.22)))) for j in range(n)] #cross flow effectiveness
        Ta_mid = [Ta_in - effect[j]*C_min*(Ta_in - Tw_mid[i])/C_air for j in range(n)]
        if Tdp<Ta_mid[-1]:
            frac[i] = 1
            UA_wet[i] = -1e8
        else:
            x_a = 1 + 4.769e-3*(Tdp - Ta_in0)
            x_w = 1 + (.014/(1 + .014*Tw_in0))*(Tw_in - Tw_in0)
            hA_a = x_a*((m_a/m_a_rated)**.8)*hA_a0
            hA_w = x_w*((m_w/m_w_rated)**.85)*hA_w0
            if Ta_in<Tdp:
                frac[i] = 0
            else:
                f1 = [0]
                f1.extend(f)
                Ta = [Ta_in]
                Ta.extend(Ta_mid)
                frac[i] = two_way_interp(Tdp,Ta,f1)
            UA_wet[i] = (1-frac[i])/(1/hA_w + 1/(n_f*hA_a))
    return UA_wet,frac


"""Utilities for handling HDF5 files.

Context managers:
h5file_context

Functions:
new_hdf5_file
write_records_to_new_hdf5_table
write_user_attributes
write_hdf5_to_csv

Classes:
DatetimeFloatConverter
"""

from contextlib import contextmanager
import csv
import datetime as dt

import numpy as np
import tables as tb

from eagers.config.path_spec import USER_DIR_HDF5_DATASETS


@contextmanager
def h5file_context(*args, **kwargs):
    """HDF5 file context that automatically closes file when it's done
    being used.

    Refer to PyTables documentation on tables.open_file() for positional
    and keyword arguments.
    """
    h5_file = tb.open_file(*args, **kwargs)
    try:
        yield h5_file
    finally:
        h5_file.close()


def new_hdf5_file(filepath, *, overwrite=False):
    """Create a new, empty HDF5 file at the specified path.

    Positional arguments:
    filepath - (pathlib.Path) Path to file.

    Keyword arguments:
    overwrite - (bool) (Default: False) Whether an existing file at the
        given path should be overwritten.
    """
    if not overwrite and filepath.exists():
        raise ValueError(f"{filepath} already exists.")
    h5f = tb.open_file(filepath, 'w')
    h5f.close()


def write_records_to_new_hdf5_table(
        filepath, tablename, records, descr=None, *, timestamp_cols=None):
    """Write records to a new table in a pre-existing HDF5 file.

    Positional arguments:
    filepath - (pathlib.Path) Path to HDF5 file.
    tablename - (str) Name of new table.
    records - (list of tuples) Records to append to the new table.

    Keyword arguments:
    descr - (dict or tables.IsDescription or numpy.dtype) (Default:
        None) Table structure description. If None, records must be a
        NumPy structured array.
    timestamp_cols - (list of str) (Default: None) List of column names
        containing timestamps. These will be converted to floats.
    """
    if descr is None:
        try:
            descr = records.dtype
        except AttributeError:
            raise ValueError(
                "If descr is not passed, records must be a NumPy structured "
                "array.")
    # Open file in "r"ead-plus mode.  This allows addition of tables to
    # an existing file only.
    with h5file_context(filepath, 'r+') as h5f:
        # Create table based on specified dimensions.
        table = h5f.create_table('/', tablename, descr)
        table.append(records)


def write_user_attributes(filepath, tablename, attributes):
    """Write the specified user attributes to the HDF5 file at the given
    path.

    Positional arguments:
    filepath - (pathlib.Path) Path to HDF5 file.
    tablename - (str) Name of table to assign the attributes to.
    attributes - (dict) Name-value pairs of attributes to be written.
    """
    # Open file in "r"ead-plus mode.  This allows writing of attributes
    # to an existing file only.
    with h5file_context(filepath, 'r+') as h5f:
        table = getattr(h5f.root, tablename)
        for name, value in attributes.items():
            setattr(table.attrs, name, value)


def write_hdf5_to_csv(input_filename, category, to_datetime, output_filepath):
    """Write HDF5 data set to CSV format.

    Positional arguments:
    input_filename - (str) Data set name, w/ file extension.
    category - (str) Data set category.
    to_datetime - (bool) Whether to convert timestamps to datetime
        format.
    output_filepath - (pathlib.Path) Path to output file.
    """
    # Read specified file.
    input_filepath = USER_DIR_HDF5_DATASETS / category / input_filename
    with h5file_context(input_filepath, mode='r') as h5file:
        table = getattr(h5file.root, category)
        raw_data = table.read()
    # Convert datetimes if needed.
    headers = raw_data.dtype.names
    if to_datetime and 'timestamp' in headers:
        # Convert values.
        raw_data['timestamp'] = DatetimeFloatConverter.f2d_arr2arr(
            raw_data['timestamp']
        )
        # Convert data type.
        new_dtype = raw_data.dtype.descr
        new_dtype[headers.index('timestamp')] = ('timestamp', 'M8[us]')
        raw_data = raw_data.astype(new_dtype)
    # Write data to CSV file.
    with open(output_filepath, 'w', newline='') as csvf:
        writer = csv.writer(csvf)
        # Write metadata.
        writer.writerow([input_filename])
        # Write headers.
        writer.writerow(headers)
        # Write data.
        for row in raw_data:
            writer.writerow(row)


class DatetimeFloatConverter:
    """Defines methods for converting between datetimes and floats.
    Since there is no way to store datetimes in HDF5 files, it is
    necessary to do this conversion in both directions - datetime to
    float for write operations and vice versa for reading. Floats
    represent the number of seconds since epoch (Unix time) (Jan 1,
    1970).

    Cf.:
    https://stackoverflow.com/questions/25706423/store-and-extract-numpy-datetimes-in-pytables
    https://stackoverflow.com/questions/6999726/how-can-i-convert-a-datetime-object-to-milliseconds-since-epoch-unix-time-in-p
    https://stackoverflow.com/questions/12137277/how-can-i-make-a-python-numpy-arange-of-datetime
    """

    @staticmethod
    def d2f_sgl2sgl(d):
        """Return single float representation of given single datetime
        as number of seconds since epoch (Jan 1, 1970).

        Positional arguments:
        d - (datetime) Datetime.
        """
        return (d - dt.datetime.utcfromtimestamp(0)) / dt.timedelta(seconds=1)

    @staticmethod
    def d2f_arr2arr(d):
        """Return NumPy float array representation of given NumPy
        datetime array as number of seconds since epoch (Jan 1, 1970).

        Positional arguments:
        d - (ndarray of datetimes) NumPy array of datetimes.
        """
        if d.dtype.kind == 'O':
            return (
                (d - dt.datetime.utcfromtimestamp(0)) / dt.timedelta(seconds=1)
            ).astype('f8')
        elif d.dtype.kind == 'M':
            # Convert to microsecond datetime64 first because conversion
            # to float differs depending on the datetime64 frequency.
            return d.astype('M8[us]').astype('f8') / 1e6

    @staticmethod
    def d2f_lst2arr(d):
        """Return NumPy float array representation of given Python
        datetime list as number of seconds since epoch (Jan 1, 1970).

        Positional arguments:
        d - (list of datetimes) List of datetimes.
        """
        return np.array([
            (x - dt.datetime.utcfromtimestamp(0)) / dt.timedelta(seconds=1)
            for x in d
        ])

    @staticmethod
    def d2f_lst2lst(d):
        """Return Python datetime list representation of given Python
        datetime list as number of seconds since epoch (Jan 1, 1970).

        Positional arguments:
        d - (list of datetimes) List of datetimes.
        """
        return [
            (x - dt.datetime.utcfromtimestamp(0)) / dt.timedelta(seconds=1)
            for x in d
        ]

    @staticmethod
    def f2d_sgl2sgl(f, dt64=False):
        """Return single datetime representation of given single float,
        which is the number of seconds since epoch (Jan 1, 1970).

        Positional arguments:
        f - (float) Float.

        Keyword arguments:
        dt64 - (bool) (Default: False) Whether to convert to a NumPy
            datetime64 object instead of datetime.datetime.
        """
        if dt64:
            return np.float64(f * 1e6).astype('M8[us]')
        return dt.datetime.utcfromtimestamp(0) + dt.timedelta(seconds=f)

    @staticmethod
    def f2d_arr2arr(f):
        """Return NumPy datetime array representation of given NumPy
        float array, which are the number of seconds since epoch (Jan 1,
        1970).

        Positional arguments:
        f - (ndarray of floats) NumPy array of floats.
        """
        return (f * 1e6).astype('M8[us]')

    @staticmethod
    def f2d_arr2lst(f):
        """Return Python datetime list representation of given NumPy
        float array, which are the number of seconds since epoch (Jan 1,
        1970).

        Positional arguments:
        f - (ndarray of floats) NumPy array of floats.
        """
        return (f * 1e6).astype('M8[us]').tolist()

    @staticmethod
    def f2d_lst2lst(f):
        """Return Python datetime list representation of given Python
        float list, which are the number of seconds since epoch (Jan 1,
        1970).

        Positional arguments:
        f - (list of floats) List of floats.
        """
        epoch = dt.datetime.utcfromtimestamp(0)
        return [epoch + dt.timedelta(seconds=x) for x in f]



"""Units-related configuration variables.
"""


OPERATOR = {
    '*': 'times',
    '-': 'times',
    '/': 'per',
}

PREFIX = {
    'k': 'kilo',
    'm': 'milli',
    'M': 'mega',
}

UNIT_ANALOG = {
    '%': 'pct',
    'acre': 'acre',
    'Acre': 'acre',
    'cfs': 'cfs',
    'ft': 'ft',
    'hr': 'hr',
    'W': 'watt',
}


# CONVERSIONS

# kWh to Btu.
# Wikipedia: British thermal unit
# Accessed: 07/09/2020
# https://en.wikipedia.org/wiki/British_thermal_unit#As_a_unit_of_power
Btu_per_kWh = 3412.142

# Cubic feet natural gas to Btu.
# U.S. E.I.A.: What are Ccf, Mcf, Btu, and therms? How do I convert
# natural gas prices in dollars per Ccf or Mcf to dollars per Btu or
# therm?
# Accessed: 07/09/2020
# https://www.eia.gov/tools/faqs/faq.php?id=45&t=8
#
# NOTE: Normally, 1037 Btu/cf natural gas would be used. This number is
# used instead to be consistent with previous results obtained with
# EAGERS.
Btu_per_Mcf_natgas = 1_000_000  # _ are thousands separators: PEP 378.

# Thousand cubic feet natural gas to kWh.
# With 1,000,000 Btu/Mcf and 3,412.142 Btu/kWh: 293.07 kWh/Mcf
kWh_per_Mcf_natgas = Btu_per_Mcf_natgas / Btu_per_kWh

# MMBtu to kWh.
kWh_per_MMBtu_natgas = 293.15


# REFERENCE VALUES

# Benchmark heat rate (Btu/kWh).
# https://www.eia.gov/todayinenergy/detail.php?id=9911
HEAT_RATE_BENCHMARK = 7_000
